from __future__ import annotations

from dataclasses import dataclass
from typing import List, Tuple, Optional, Union, Sequence

import numpy as np
from numpy.random import SeedSequence

from .base_oscillation_kind import BaseOscillationKind


class GenerationContext:
    def __init__(self, seed: SeedSequence):
        self.seed: SeedSequence = seed
        self.rng: np.random.Generator = np.random.default_rng(self.seed)

    def to_bo(self, channel: int = 0, previous_channels: Sequence[np.ndarray] = ()) -> BOGenerationContext:
        return BOGenerationContext(
            seed=self.seed,
            rng=self.rng,
            channel=channel,
            previous_channels=list(previous_channels)
        )

    def to_anomaly(self, bo: 'BaseOscillationInterface',  # type: ignore  # to prevent circular import
                   previous_anomaly_positions: Sequence[Tuple[int, int]]) -> AnomalyGenerationContext:
        return AnomalyGenerationContext(
            seed=self.seed,
            rng=self.rng,
            base_oscillation=bo,
            previous_anomaly_positions=list(previous_anomaly_positions)
        )

    @staticmethod
    def re_seed(new_seeds: Union[int, Optional[int], Sequence[int]],
                base_seed: Union[int, SeedSequence] = 0) -> SeedSequence:
        if isinstance(base_seed, SeedSequence):
            if isinstance(base_seed.entropy, int):
                initial_entropy: Sequence[int] = [base_seed.entropy]
            elif base_seed.entropy is not None:
                initial_entropy = base_seed.entropy
            else:
                initial_entropy = []
        else:
            initial_entropy = [base_seed]
        if new_seeds is None:
            new_seeds = []
        elif isinstance(new_seeds, int):
            new_seeds = [new_seeds]
        return SeedSequence(initial_entropy + new_seeds)  # type: ignore  # wrong type declaration in SeedSequence


@dataclass
class BOGenerationContext(GenerationContext):
    seed: SeedSequence
    rng: np.random.Generator
    channel: int
    previous_channels: List[np.ndarray]
    is_trend: bool = False

    def to_trend(self) -> BOGenerationContext:
        return BOGenerationContext(
            seed=self.seed,
            rng=self.rng,
            channel=self.channel,
            previous_channels=self.previous_channels,
            is_trend=True
        )


@dataclass
class AnomalyGenerationContext(GenerationContext):
    seed: SeedSequence
    rng: np.random.Generator
    base_oscillation: 'BaseOscillationInterface'  # type: ignore  # to prevent circular import
    previous_anomaly_positions: List[Tuple[int, int]]

    @property
    def timeseries_periods(self) -> Optional[int]:
        return self.base_oscillation.get_timeseries_periods()

    @property
    def base_oscillation_kind(self) -> BaseOscillationKind:
        return self.base_oscillation.get_base_oscillation_kind()



import os
from enum import Enum
from typing import List, Dict, Optional, Iterable, Any, Union

import numpy as np
import pandas as pd

from gutenTAG.addons import BaseAddOn, AddOnProcessContext, AddOnFinalizeContext
from gutenTAG.generator import TimeSeries
from gutenTAG.utils.default_values import default_values
from gutenTAG.utils.global_variables import SUPERVISED_FILENAME, UNSUPERVISED_FILENAME, SEMI_SUPERVISED_FILENAME, \
    BASE_OSCILLATIONS, ANOMALIES, PARAMETERS, BASE_OSCILLATION, BASE_OSCILLATION_NAMES


columns = [
    "collection_name",
    "dataset_name",
    "train_path",
    "test_path",
    "dataset_type",
    "datetime_index",
    "split_at",
    "train_type",
    "train_is_normal",
    "input_type",
    "length",
    "dimensions",
    "contamination",
    "num_anomalies",
    "min_anomaly_length",
    "median_anomaly_length",
    "max_anomaly_length",
    "mean",
    "stddev",
    "trend",
    "stationarity",
    "period_size"
]


class LearningType(Enum):
    Unsupervised = "unsupervised"
    Supervised = "supervised"
    SemiSupervised = "semi-supervised"

    def get_filename(self) -> Optional[str]:
        if self == LearningType.Supervised:
            return SUPERVISED_FILENAME
        elif self == LearningType.SemiSupervised:
            return SEMI_SUPERVISED_FILENAME
        return None


class TimeEvalAddOn(BaseAddOn):
    def __init__(self):
        self.df = pd.DataFrame(columns=columns)
        self.key = self.__class__.__name__

    def process(self, ctx: AddOnProcessContext) -> AddOnProcessContext:
        ts = ctx.timeseries
        config = ctx.config
        datasets = [
            self._process_timeseries(config, ts, LearningType.Unsupervised)
        ]
        if ts.supervised:
            datasets.append(self._process_timeseries(config, ts, LearningType.Supervised))
        if ts.semi_supervised:
            datasets.append(self._process_timeseries(config, ts, LearningType.SemiSupervised))
        return ctx.store_data(self.key, {
            "name": ts.dataset_name,
            "datasets": datasets
        })

    def finalize(self, ctx: AddOnFinalizeContext) -> None:
        df = pd.DataFrame(columns=columns)
        # add metadata
        for ts_obj in ctx.get_data(self.key):
            df = df.append(ts_obj["datasets"], ignore_index=True)
        self._set_global_vals(df)
        self.df = df
        if ctx.should_save and ctx.output_folder is not None:
            filename = os.path.join(ctx.output_folder, "datasets.csv")
            df.to_csv(filename, index=False)

    def _process_timeseries(self, config: Dict, generator: TimeSeries, tpe: LearningType) -> Dict[str, Any]:
        dataset: Dict[str, Any] = dict()

        dataset_name = generator.dataset_name
        filename = tpe.get_filename()
        if filename is not None:
            dataset["train_path"] = f"{dataset_name}/{filename}"

        ts = generator.timeseries
        assert ts is not None, "Timeseries should not be None!"

        dataset["dataset_name"] = f"{dataset_name}.{tpe.value}"
        dataset["test_path"] = f"{dataset_name}/{UNSUPERVISED_FILENAME}"
        dataset["input_type"] = "univariate" if ts.shape[1] == 1 else "multivariate"
        dataset["length"] = config.get(PARAMETERS.LENGTH, 10000)
        dataset["dimensions"] = ts.shape[1]
        dataset["contamination"] = self._calc_contamination(config.get(ANOMALIES, []), dataset[PARAMETERS.LENGTH])
        dataset["num_anomalies"] = len(config.get(ANOMALIES, []))
        dataset["min_anomaly_length"] = min([anomaly.get("length") for anomaly in config.get(ANOMALIES, [])])
        dataset["median_anomaly_length"] = np.median([anomaly.get(PARAMETERS.LENGTH) for anomaly in config.get(ANOMALIES, [])])
        dataset["max_anomaly_length"] = max([anomaly.get(PARAMETERS.LENGTH) for anomaly in config.get(ANOMALIES, [])])
        dataset["train_type"] = tpe.value
        dataset["train_is_normal"] = False if tpe == LearningType.Supervised else True
        dataset["mean"] = ts.mean()
        dataset["stddev"] = ts.std(axis=1).mean()
        dataset["trend"] = config.get(BASE_OSCILLATION, {}).get(PARAMETERS.TREND, {}).get(PARAMETERS.KIND, np.NAN)
        dataset["period_size"] = TimeEvalAddOn._calc_period_size(config.get(BASE_OSCILLATION, config.get(BASE_OSCILLATIONS, [{}])), dataset[PARAMETERS.LENGTH])
        return dataset

    @staticmethod
    def _set_global_vals(df: pd.DataFrame) -> None:
        df["collection_name"] = "GutenTAG"
        df["dataset_type"] = "synthetic"
        df["datetime_index"] = False
        df["split_at"] = np.NAN
        df["stationarity"] = np.NAN

    @staticmethod
    def _calc_contamination(anomalies: Iterable[Dict], ts_length: int) -> float:
        anomaly_lengths = [anomaly.get(PARAMETERS.LENGTH, default_values[ANOMALIES][PARAMETERS.LENGTH]) for anomaly in anomalies]
        if len(anomaly_lengths) > 0:
            return sum(anomaly_lengths) / ts_length
        return 0

    @staticmethod
    def _calc_period_size(base: Union[Dict[str, Any], List[Dict[str, Any]]], length: int) -> float:
        bases: List[Dict[str, Any]] = []
        if type(base) == dict:
            bases.append(base)  # type: ignore  # does not understand the condition before
        elif type(base) == list:
            bases = base  # type: ignore  # does not understand the condition before

        periods = []

        for dim in bases:
            frequency = dim.get(PARAMETERS.FREQUENCY)
            kind = dim.get(PARAMETERS.KIND)
            if frequency is None or kind not in [BASE_OSCILLATION_NAMES.SINE, BASE_OSCILLATION_NAMES.ECG, BASE_OSCILLATION_NAMES.RANDOM_MODE_JUMP]:
                periods.append(np.NAN)
            elif kind in [BASE_OSCILLATION_NAMES.SINE, BASE_OSCILLATION_NAMES.ECG]:
                periods.append(int(100 / frequency))
            elif kind == BASE_OSCILLATION_NAMES.RANDOM_MODE_JUMP:
                periods.append(int(length / frequency))
        return float(np.nanmedian(periods))



from .global_variables import BASE_OSCILLATIONS, ANOMALIES, PARAMETERS


default_values = {
    BASE_OSCILLATIONS: {
        PARAMETERS.LENGTH: 10000,
        PARAMETERS.KIND: "sine",
        PARAMETERS.FREQUENCY: 10.0,
        PARAMETERS.AMPLITUDE: 1.0,
        PARAMETERS.VARIANCE: 0.0,
        PARAMETERS.AVG_PATTERN_LENGTH: 10,
        PARAMETERS.VARIANCE_PATTERN_LENGTH: 0.0,
        PARAMETERS.VARIANCE_AMPLITUDE: 2.0,
        PARAMETERS.FREQ_MOD: 0.0,
        PARAMETERS.POLYNOMIAL: [1, 1],
        PARAMETERS.TREND: None,
        PARAMETERS.OFFSET: 0.0,
        PARAMETERS.SMOOTHING: 0.01,
        PARAMETERS.CHANNEL_DIFF: 0.0,
        PARAMETERS.CHANNEL_OFFSET: 1.0,
        PARAMETERS.RANDOM_SEED: None,
        PARAMETERS.FORMULA: None,
        PARAMETERS.ECG_SIM_METHOD: "simple"
    },
    ANOMALIES: {
        PARAMETERS.LENGTH: 200,
        PARAMETERS.POSITION: "middle",
        PARAMETERS.CHANNEL: 0,
        PARAMETERS.CREEP_LENGTH: 0
    }
}



import logging


class GutenTagLogger:
    def __init__(self):
        self.logger = logging.getLogger("GutenTAG")
        self.logger.setLevel(logging.DEBUG)
        self.logger.addHandler(logging.StreamHandler())

    def warn_false_combination(self, anomaly: str, base_oscillation: str):
        self.logger.warning(f"You tried to generate '{anomaly}' on '{base_oscillation}'. That doesn't work! Guten Tag!")


if __name__ == "__main__":
    logger = GutenTagLogger()
    logger.warn_false_combination("test", "test2")



import logging
from dataclasses import dataclass
from typing import Type

from . import BaseAnomaly
from .. import AnomalyProtocol
from ...utils.base_oscillation_kind import BaseOscillationKind


@dataclass
class AnomalyModeCorrelationParameters:
    pass


class AnomalyModeCorrelation(BaseAnomaly):
    def __init__(self, parameters: AnomalyModeCorrelationParameters):
        super().__init__()

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        base = anomaly_protocol.base_oscillation
        if anomaly_protocol.base_oscillation_kind == BaseOscillationKind.RandomModeJump:
            timeseries = base.timeseries
            subsequence = timeseries[anomaly_protocol.start:anomaly_protocol.end] * -1
            anomaly_protocol.subsequences.append(subsequence)
        else:
            logging.warning("A `mode_correlation` anomaly can be injected in only a `random_mode_jump` base oscillation!")
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return True

    @staticmethod
    def get_parameter_class() -> Type[AnomalyModeCorrelationParameters]:
        return AnomalyModeCorrelationParameters



from dataclasses import dataclass
from typing import Type

import numpy as np

from . import BaseAnomaly
from .. import AnomalyProtocol


@dataclass
class AnomalyExtremumParameters:
    min: bool = False
    local: bool = False
    context_window: int = 10


class AnomalyExtremum(BaseAnomaly):
    def __init__(self, parameters: AnomalyExtremumParameters):
        super().__init__()
        self.min = parameters.min
        self.local = parameters.local
        self.context_window = parameters.context_window

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        length = anomaly_protocol.end - anomaly_protocol.start
        if length != 1:
            self.logger.logger.warn(f"Extremum anomaly can only have a length of 1 (was set to {length})! Ignoring.")
            anomaly_protocol.end = anomaly_protocol.start + 1
            anomaly_protocol.labels.length = 1

        base: np.ndarray = anomaly_protocol.base_oscillation.timeseries
        if self.local:
            context_start = max(anomaly_protocol.start - self.context_window//2, 0)
            context_end = min(anomaly_protocol.end + self.context_window//2, base.shape[0])
            context = base[context_start:context_end]
            diff = context.max() - context.min()
            extremum = anomaly_protocol.rng.random() * diff
        else:
            diff = base.max() - base.min()
            extremum = (anomaly_protocol.rng.random() + 0.5) * diff

        # let extremum be significant enough to be distinguishable from noise
        max_noise: float = np.max(np.abs(anomaly_protocol.base_oscillation.noise))
        if extremum < 2*max_noise:
            extremum += 2*max_noise

        if self.min:
            value = base[anomaly_protocol.start] - extremum
        else:
            value = base[anomaly_protocol.start] + extremum
        anomaly_protocol.subsequences.append(np.array([value]))
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return False

    @staticmethod
    def get_parameter_class() -> Type[AnomalyExtremumParameters]:
        return AnomalyExtremumParameters



from dataclasses import dataclass
from typing import Type
import numpy as np

from . import BaseAnomaly
from .. import AnomalyProtocol


@dataclass
class AnomalyMeanParameters:
    offset: float = 0.0


class AnomalyMean(BaseAnomaly):
    def __init__(self, parameters: AnomalyMeanParameters):
        super().__init__()
        self.offset = parameters.offset

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        base = anomaly_protocol.base_oscillation
        ts: np.ndarray = base.timeseries
        creep = self.generate_creep(anomaly_protocol)
        subsequence = ts[anomaly_protocol.start:anomaly_protocol.end] + self.offset * creep
        anomaly_protocol.subsequences.append(subsequence)
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return False

    @staticmethod
    def get_parameter_class() -> Type[AnomalyMeanParameters]:
        return AnomalyMeanParameters



from typing import Optional

import numpy as np
from scipy.stats import norm
from sklearn.preprocessing import MinMaxScaler

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext


def _gen_steps(ctx: BOGenerationContext, length: int) -> np.ndarray:
    steps = ctx.rng.choice([-1., 0., 1.], size=length - 1)
    return np.r_[0, steps].cumsum()


class RandomWalk(BaseOscillationInterface):
    def get_timeseries_periods(self) -> Optional[int]:
        return None

    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.RandomWalk

    def generate_only_base(self,
                           ctx: BOGenerationContext,
                           length: Optional[int] = None,
                           amplitude: Optional[float] = None,
                           smoothing: Optional[float] = None,
                           *args, **kwargs) -> np.ndarray:
        length = length or self.length
        amplitude = amplitude or self.amplitude
        smoothing = smoothing or self.smoothing

        if smoothing:
            filter_size = int(smoothing * length)
            ts = _gen_steps(ctx, length + filter_size - 1)
            gaussian = norm.pdf(np.linspace(-1.5, 1.5, filter_size))
            ts_filter = gaussian / gaussian.sum()
            ts = np.convolve(ts, ts_filter, "valid")
        else:
            ts = _gen_steps(ctx, length)

        return MinMaxScaler(feature_range=[-amplitude, amplitude]).fit_transform(ts.reshape(-1, 1)).reshape(-1)



from typing import List, Optional, Tuple

import numpy as np

from ...anomalies import AnomalyProtocol, LabelRange, Anomaly
from ...base_oscillations.interface import BaseOscillationInterface
from ...utils.types import GenerationContext


class Consolidator:
    def __init__(self,
                 base_oscillations: List[BaseOscillationInterface],
                 anomalies: List[Anomaly],
                 random_seed: Optional[int] = None):
        self.consolidated_channels: List[BaseOscillationInterface] = base_oscillations
        self.anomalies: List[Anomaly] = anomalies
        self.generated_anomalies: List[Tuple[AnomalyProtocol, int]] = []
        self.timeseries: Optional[np.ndarray] = None
        self.labels: Optional[np.ndarray] = None
        self.random_seed: Optional[int] = random_seed

    def add_channel(self, channel: BaseOscillationInterface):
        self.consolidated_channels.append(channel)

    def get_channel(self, channel: int) -> BaseOscillationInterface:
        return self.consolidated_channels[channel]

    def generate(self, ctx: GenerationContext) -> Tuple[np.ndarray, np.ndarray]:
        channels: List[np.ndarray] = []
        for c, bo in enumerate(self.consolidated_channels):
            bo.generate_timeseries_and_variations(ctx.to_bo(c, channels))  # type: ignore  # timeseries gets set in generate_timeseries_and_variations()
            if bo.timeseries is not None:
                channels.append(bo.timeseries)
        self.timeseries = self._stack_channels(channels)
        self.labels = np.zeros(self.timeseries.shape[0], dtype=np.int8)
        self.generate_anomalies(ctx)

        self.apply_anomalies()
        self.apply_variations()
        return self.timeseries, self.labels

    def apply_variations(self):
        for c, bo in enumerate(self.consolidated_channels):
            self.timeseries[:, c] += bo.noise + bo.trend_series + bo.offset

    def apply_anomalies(self):
        label_ranges: List[LabelRange] = []
        for (protocol, channel) in self.generated_anomalies:
            if len(protocol.subsequences) > 0:
                subsequence = np.vstack(protocol.subsequences).sum(axis=0)
                self.timeseries[protocol.start:protocol.end, channel] = subsequence
            label_ranges.append(protocol.labels)

        self._add_label_ranges_to_labels(label_ranges)

    def generate_anomalies(self, ctx: GenerationContext):
        positions: List[Tuple[int, int]] = []
        for anomaly in self.anomalies:
            current_base_oscillation = self.consolidated_channels[anomaly.channel]
            anomaly_protocol = anomaly.generate(ctx.to_anomaly(current_base_oscillation, positions))
            positions.append((anomaly_protocol.start, anomaly_protocol.end))
            self.generated_anomalies.append((anomaly_protocol, anomaly.channel))

    def _stack_channels(self, channels: List[np.ndarray]) -> np.ndarray:
        assert all([len(x.shape) == 1 for x in channels]), "The resulting channels have the wrong shape. Correct shape: `(l, d)`."
        return np.vstack(channels).transpose()

    def _add_label_ranges_to_labels(self, label_ranges: List[LabelRange]):
        if self.labels is not None:
            for label_range in label_ranges:
                self.labels[label_range.start:label_range.start + label_range.length] = 1
        else:
            raise AssertionError("You cannot run this method before initializing the `labels` field!")



from enum import Enum


class BaseOscillationKind(Enum):
    Sine = "sine"
    RandomWalk = "random_walk"
    CylinderBellFunnel = "cylinder_bell_funnel"
    ECG = "ecg"
    Polynomial = "polynomial"
    RandomModeJump = "random_mode_jump"
    Formula = "formula"



from dataclasses import dataclass
from typing import Type

from . import BaseAnomaly
from .. import AnomalyProtocol
from ...utils.base_oscillation_kind import BaseOscillationKind


@dataclass
class AnomalyVarianceParameters:
    variance: float = 0.0


class AnomalyVariance(BaseAnomaly):
    def __init__(self, parameters: AnomalyVarianceParameters):
        super().__init__()
        self.variance = parameters.variance

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        base = anomaly_protocol.base_oscillation
        if anomaly_protocol.base_oscillation_kind == BaseOscillationKind.CylinderBellFunnel:
            subsequence = base.generate_only_base(anomaly_protocol.ctx.to_bo(), variance=self.variance)[anomaly_protocol.start:anomaly_protocol.end]
            anomaly_protocol.subsequences.append(subsequence)
        else:
            length = anomaly_protocol.end - anomaly_protocol.start
            variance_diff = self.variance - base.variance
            creep = self.generate_creep(anomaly_protocol) * variance_diff + base.variance  # from 0 to variance_diff
            creep /= self.variance * base.amplitude  # get relative transition from base variance to anomaly variance
            subsequence_noise = base.generate_noise(anomaly_protocol.ctx.to_bo(), self.variance * base.amplitude, length)
            base.noise[anomaly_protocol.start:anomaly_protocol.end] = subsequence_noise * creep
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return False

    @staticmethod
    def get_parameter_class() -> Type[AnomalyVarianceParameters]:
        return AnomalyVarianceParameters



from __future__ import annotations

import numpy as np

from enum import Enum
from typing import List, Optional, Tuple

from .types import AnomalyProtocol, BaseAnomaly, LabelRange
from .types.kind import AnomalyKind
from ..utils.types import AnomalyGenerationContext


class Position(Enum):
    Beginning = "beginning"
    Middle = "middle"
    End = "end"

    @property
    def id(self) -> int:
        if self == Position.Beginning:
            return 0
        elif self == Position.Middle:
            return 1
        elif self == Position.End:
            return 2
        else:
            raise ValueError(f"The position '{self}' is not yet supported! Guten Tag!")


class Anomaly:
    """This class acts like a generator graph, that collects all options in the beginning and,
    while being injected to a BaseOscillation, performs the changes."""

    def __init__(self,
                 position: Position,
                 exact_position: Optional[int],
                 anomaly_length: int,
                 channel: int = 0,
                 creep_length: int = 0):
        self.position = position
        self.exact_position = exact_position
        self.anomaly_length = anomaly_length
        self.channel = channel
        self.creep_length = creep_length

        self.anomaly_kinds: List[BaseAnomaly] = []
        self._requires_period_start_position: bool = False

    def set_anomaly(self, anomaly_kind: BaseAnomaly) -> Anomaly:
        self.anomaly_kinds.append(anomaly_kind)
        self._requires_period_start_position = self._requires_period_start_position or anomaly_kind.requires_period_start_position
        return self

    def generate(self, ctx: AnomalyGenerationContext) -> AnomalyProtocol:
        if self.exact_position is None:
            start, end = self._find_position(ctx)
        else:
            start, end = self.exact_position, self.exact_position + self.anomaly_length

        length = end - start
        label_range = LabelRange(start, length)
        protocol = AnomalyProtocol(start, end, self.channel, ctx, label_range, creep_length=self.creep_length)

        for anomaly in self.anomaly_kinds:
            protocol = anomaly.generate(protocol)

        return protocol

    def _find_position(self, ctx: AnomalyGenerationContext, max_tries: int = 50) -> Tuple[int, int]:
        n_tries = max_tries
        while n_tries > 0:
            pos = self._get_random_position(ctx)
            n_tries -= 1
            if pos[1] < ctx.base_oscillation.length and not self._has_collision(pos, ctx.previous_anomaly_positions):
                return pos
        raise ValueError(f"Giving up on finding a position for {self.anomaly_length}-point anomaly at {self.position} "
                         f"in channel {self.channel}! Maximum number of retries ({max_tries}) exceeded!")

    def _get_random_position(self, ctx: AnomalyGenerationContext) -> Tuple[int, int]:
        timeseries_length = ctx.base_oscillation.length
        timeseries_periods = ctx.timeseries_periods
        if (
                not self._requires_period_start_position
                or timeseries_periods is None
                or timeseries_periods <= 6
        ):
            return self._get_random_position_no_periodicity(ctx)

        start_period = 0
        period_size = timeseries_length // timeseries_periods
        periods_per_section = timeseries_periods // 3
        if periods_per_section > 1:
            start_period = ctx.rng.choice(list(range(periods_per_section)))

        position = self.position.id
        start = period_size * (position*periods_per_section + start_period)
        end = start + self.anomaly_length
        return start, end

    def _get_random_position_no_periodicity(self, ctx: AnomalyGenerationContext) -> Tuple[int, int]:
        timeseries_length = ctx.base_oscillation.length
        section_size = timeseries_length // 3
        position_in_section = ctx.rng.choice(np.arange(section_size))

        position = self.position.id
        start = (position * section_size) + position_in_section
        end = start + self.anomaly_length
        return start, end

    @staticmethod
    def _has_collision(current_pos: Tuple[int, int], other_pos: List[Tuple[int, int]]) -> bool:
        if len(other_pos) == 0:
            return False
        others = np.array(other_pos)
        start_collision = (others[:, 0] <= current_pos[0]) & (current_pos[0] <= others[:, 1])
        end_colllision = (others[:, 0] <= current_pos[1]) & (current_pos[1] <= others[:, 1])
        return bool(np.any(start_collision | end_colllision))



from typing import Optional

import neurokit2 as nk
import numpy as np

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext

# we fix the sampling rate to 100 points = 1s
sampling_rate = 100


class ECG(BaseOscillationInterface):
    def get_timeseries_periods(self) -> Optional[int]:
        return int((self.length // sampling_rate) * (self.frequency / 100 * sampling_rate))

    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.ECG

    def generate_only_base(self,
                           ctx: BOGenerationContext,
                           length: Optional[int] = None,
                           frequency: Optional[float] = None,
                           amplitude: Optional[float] = None,
                           *args, **kwargs) -> np.ndarray:
        length = length or self.length
        frequency = frequency or self.frequency
        amplitude = amplitude or self.amplitude

        duration = length // sampling_rate
        # frequency = beats per 100 points = beats per second
        heart_rate = frequency / 100 * sampling_rate * 60

        ecg = nk.ecg_simulate(duration=duration,
                              sampling_rate=sampling_rate,
                              heart_rate=heart_rate,
                              length=length,
                              random_state=ctx.rng.integers(0, int(1e9)),
                              noise=0,
                              method=self.ecg_sim_method)
        return ecg * amplitude



from pathlib import Path


SUPERVISED_FILENAME = "train_anomaly.csv"
SEMI_SUPERVISED_FILENAME = "train_no_anomaly.csv"
UNSUPERVISED_FILENAME = "test.csv"


BASE_OSCILLATION = "base-oscillation"
BASE_OSCILLATIONS = BASE_OSCILLATION + "s"
TIMESERIES = "timeseries"
ANOMALIES = "anomalies"


class BASE_OSCILLATION_NAMES:
    SINE = "sine"
    RANDOM_WALK = "random-walk"
    CYLINDER_BELL_FUNNEL = "cylinder-bell-funnel"
    ECG = "ecg"
    POLYNOMIAL = "polynomial"
    RANDOM_MODE_JUMP = "random-mode-jump"
    FORMULA = "formula"


class ANOMALY_TYPE_NAMES:
    AMPLITUDE = "amplitude"
    EXTREMUM = "extremum"
    FREQUENCY = "frequency"
    MEAN = "mean"
    PATTERN = "pattern"
    PATTERN_SHIFT = "pattern-shift"
    PLATFORM = "platform"
    TREND = "trend"
    VARIANCE = "variance"
    MODE_CORRELATION = "mode-correlation"


class PARAMETERS:
    OSCILLATION = "oscillation"
    LENGTH = "length"
    FREQUENCY = "frequency"
    AMPLITUDE = "amplitude"
    VARIANCE = "variance"
    AVG_PATTERN_LENGTH = "avg-pattern-length"
    VARIANCE_PATTERN_LENGTH = "variance-pattern-length"
    VARIANCE_AMPLITUDE = "variance-amplitude"
    FREQ_MOD = "freq-mod"
    POLYNOMIAL = "polynomial"
    TREND = "trend"
    OFFSET = "offset"
    SMOOTHING = "smoothing"
    CHANNEL_DIFF = "channel-diff"
    CHANNEL_OFFSET = "channel-offset"
    RANDOM_SEED = "random-seed"
    FORMULA = "formula"
    KIND = "kind"
    KINDS = "kinds"
    NAME = "name"
    CHANNELS = "channels"
    CHANNEL = "channel"
    POSITION = "position"
    EXACT_POSITION = "exact-position"
    CREEP_LENGTH = "creep-length"
    ECG_SIM_METHOD = "ecg-sim-method"


class CONFIG_SCHEMA:
    BASE_ID = "guten-tag-generation-config.schema.yaml"
    SCHEMA_DOMAIN = "https://example.com"
    SCHEMA_FOLDER_PATH = Path("generation-config-schema")

    SCHEMA_PART_IDS = [
        "anomaly.guten-tag-generation-config.schema.yaml",
        "anomaly-kind.guten-tag-generation-config.schema.yaml",
        "formula.guten-tag-generation-config.schema.yaml",
        "oscillation.guten-tag-generation-config.schema.yaml"
    ]

    @staticmethod
    def schema_path(schema_id: str, path: Path = SCHEMA_FOLDER_PATH) -> Path:
        return path / schema_id

    @staticmethod
    def schema_name(schema_id: str, domain: str = SCHEMA_DOMAIN) -> str:
        return f"{domain}/{schema_id}"



import contextlib
from typing import Generator

import joblib
from joblib.parallel import BatchCompletionCallBack
from tqdm import tqdm


@contextlib.contextmanager
def tqdm_joblib(tqdm_object: tqdm) -> Generator[tqdm, None, None]:
    """Context manager to patch joblib to report into tqdm progress bar given as argument.

    Directly taken from https://stackoverflow.com/a/58936697.

    Examples
    --------
    >>> import time
    >>> from joblib import Parallel, delayed
    >>>
    >>> def some_method(wait_time):
    >>>     time.sleep(wait_time)
    >>>
    >>> with tqdm_joblib(tqdm(desc="Sleeping method", total=10)):
    >>>     Parallel(n_jobs=2)(delayed(some_method)(0.2) for i in range(10))
    """

    class TqdmBatchCompletionCallback(BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()



from dataclasses import dataclass
from typing import Type

import numpy as np
from scipy.stats import norm
from sklearn.preprocessing import MinMaxScaler

from . import BaseAnomaly, AnomalyProtocol
from ...utils.base_oscillation_kind import BaseOscillationKind


@dataclass
class AnomalyAmplitudeParameters:
    amplitude_factor: float = 1.0


class AnomalyAmplitude(BaseAnomaly):
    def __init__(self, parameters: AnomalyAmplitudeParameters):
        super().__init__()
        self.amplitude_factor = parameters.amplitude_factor

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        if anomaly_protocol.base_oscillation_kind in [BaseOscillationKind.Polynomial, BaseOscillationKind.Formula, BaseOscillationKind.RandomModeJump]:
            self.logger.warn_false_combination(self.__class__.__name__, anomaly_protocol.base_oscillation_kind.name)
            return anomaly_protocol

        length = anomaly_protocol.end - anomaly_protocol.start
        if anomaly_protocol.creep_length == 0:
            transition_length = int(length * 0.2)
            plateau = np.ones(int(length * 0.6))
            start_transition = norm.pdf(np.linspace(-3, 0, transition_length), scale=1.05)
            end_transition = norm.pdf(np.linspace(0, 3, transition_length), scale=1.05)
            amplitude_bell = np.concatenate([start_transition / start_transition.max(), plateau, end_transition / end_transition.max()])
        else:
            anomaly_length = length - anomaly_protocol.creep_length
            creep = self.generate_creep(anomaly_protocol, custom_anomaly_length=int(anomaly_length * 0.8))
            end_transition = norm.pdf(np.linspace(0, 3, int(anomaly_length * 0.2)), scale=1.05)
            amplitude_bell = np.concatenate([creep, end_transition / end_transition.max()])
        if self.amplitude_factor < 1.0:
            amplitude_bell = MinMaxScaler(feature_range=(1.0, 2.0 - self.amplitude_factor)).fit_transform(amplitude_bell.reshape(-1, 1)).reshape(-1)
            amplitude_bell = amplitude_bell * -1 + 2
        else:
            amplitude_bell = MinMaxScaler(feature_range=(1.0, self.amplitude_factor)).fit_transform(amplitude_bell.reshape(-1, 1)).reshape(-1)

        subsequence = anomaly_protocol.base_oscillation.timeseries[anomaly_protocol.start:anomaly_protocol.end] * amplitude_bell
        anomaly_protocol.subsequences.append(subsequence)
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return False

    @staticmethod
    def get_parameter_class() -> Type[AnomalyAmplitudeParameters]:
        return AnomalyAmplitudeParameters



from typing import Optional, List

import numpy as np

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext


class Polynomial(BaseOscillationInterface):
    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.Polynomial

    def get_timeseries_periods(self) -> Optional[int]:
        return None

    def generate_only_base(self,
                           ctx: BOGenerationContext,
                           length: Optional[int] = None,
                           polynomial: Optional[List[float]] = None,
                           *args, **kwargs) -> np.ndarray:
        length = length or self.length
        polynomial = polynomial or self.polynomial

        base_ts = np.polynomial.Polynomial(polynomial)(np.arange(length))
        return base_ts



from enum import Enum
from typing import Dict

from .amplitude import AnomalyAmplitude
from .extremum import AnomalyExtremum
from .frequency import AnomalyFrequency
from .mean import AnomalyMean
from .mode_correlation import AnomalyModeCorrelation
from .pattern import AnomalyPattern
from .pattern_shift import AnomalyPatternShift
from .platform import AnomalyPlatform
from .trend import AnomalyTrend
from .variance import AnomalyVariance
from .. import BaseAnomaly
from ...utils.global_variables import ANOMALY_TYPE_NAMES


class AnomalyKind(Enum):
    Extremum = ANOMALY_TYPE_NAMES.EXTREMUM
    Frequency = ANOMALY_TYPE_NAMES.FREQUENCY
    Mean = ANOMALY_TYPE_NAMES.MEAN
    Pattern = ANOMALY_TYPE_NAMES.PATTERN
    PatternShift = ANOMALY_TYPE_NAMES.PATTERN_SHIFT
    Platform = ANOMALY_TYPE_NAMES.PLATFORM
    Variance = ANOMALY_TYPE_NAMES.VARIANCE
    Amplitude = ANOMALY_TYPE_NAMES.AMPLITUDE
    Trend = ANOMALY_TYPE_NAMES.TREND
    ModeCorrelation = ANOMALY_TYPE_NAMES.MODE_CORRELATION

    def create(self, parameters: Dict) -> BaseAnomaly:
        if self == AnomalyKind.Platform:
            anomaly: BaseAnomaly = AnomalyPlatform(AnomalyPlatform.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Frequency:
            anomaly = AnomalyFrequency(AnomalyFrequency.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Extremum:
            anomaly = AnomalyExtremum(AnomalyExtremum.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Variance:
            anomaly = AnomalyVariance(AnomalyVariance.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Mean:
            anomaly = AnomalyMean(AnomalyMean.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Pattern:
            anomaly = AnomalyPattern(AnomalyPattern.get_parameter_class()(**parameters))
        elif self == AnomalyKind.PatternShift:
            anomaly = AnomalyPatternShift(AnomalyPatternShift.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Amplitude:
            anomaly = AnomalyAmplitude(AnomalyAmplitude.get_parameter_class()(**parameters))
        elif self == AnomalyKind.Trend:
            anomaly = AnomalyTrend(AnomalyTrend.get_parameter_class()(**parameters))
        elif self == AnomalyKind.ModeCorrelation:
            anomaly = AnomalyModeCorrelation(AnomalyModeCorrelation.get_parameter_class()())
        else:
            raise ValueError(f"AnomalyKind {self.value} is not supported, yet! Guten Tag!")

        return anomaly

    @classmethod
    def has_value(cls, v: str) -> bool:
        values = set(item.value for item in cls)
        return v in values



from __future__ import annotations

import logging
from copy import deepcopy
from dataclasses import dataclass, asdict
from typing import Dict, Optional, Tuple, List

from ..anomalies import Anomaly, Position, AnomalyKind, BaseAnomaly
from ..base_oscillations import BaseOscillation, BaseOscillationInterface
from ..utils.compatibility import Compatibility
from ..utils.default_values import default_values
from ..utils.global_variables import BASE_OSCILLATION, BASE_OSCILLATIONS, TIMESERIES, PARAMETERS, ANOMALIES


@dataclass
class GenerationOptions:
    semi_supervised: bool = False
    supervised: bool = False
    dataset_name: Optional[str] = None

    def to_dict(self):
        return asdict(self)

    @staticmethod
    def from_dict(d: Dict) -> GenerationOptions:
        return GenerationOptions(
            semi_supervised=d.get("semi-supervised", False),
            supervised=d.get("supervised", False)
        )


ResultType = List[Tuple[List[BaseOscillationInterface], List[Anomaly], GenerationOptions, Dict]]


def decode_trend_obj(trend: Dict, length_overwrite: int) -> Optional[BaseOscillationInterface]:
    trend_key = trend.get(PARAMETERS.KIND, None)
    trend[PARAMETERS.LENGTH] = length_overwrite
    if PARAMETERS.TREND in trend:
        trend[PARAMETERS.TREND] = decode_trend_obj(trend[PARAMETERS.TREND], length_overwrite)
    return BaseOscillation.from_key(trend_key, **trend) if trend_key else None


class ConfigParser:
    def __init__(self, only: Optional[str] = None, skip_errors: bool = False):
        self.result: ResultType = []
        self.only = only
        self.raw_ts_configs: List[Dict] = []
        self.skip_errors = skip_errors

    def parse(self, config: Dict) -> ResultType:
        for i, ts in enumerate(config.get(TIMESERIES, [])):
            name = ts.get(PARAMETERS.NAME, f"ts_{i}")

            if self._skip_name(name) or not self._check_compatibility(ts):
                continue

            raw_ts_config = deepcopy(ts)
            self.raw_ts_configs.append(raw_ts_config)

            generation_options = GenerationOptions.from_dict(ts)
            generation_options.dataset_name = name

            base_oscillations = self._build_base_oscillations(ts)
            anomalies = self._build_anomalies(ts)

            self.result.append(
                (base_oscillations, anomalies, generation_options, raw_ts_config)
            )

        return self.result

    def _check_compatibility(self, ts: Dict) -> bool:
        base_oscillations = ts.get(BASE_OSCILLATIONS, [ts.get(BASE_OSCILLATION)] * ts.get(PARAMETERS.CHANNELS, 0))
        anomalies = ts.get(ANOMALIES, [])
        for anomaly in anomalies:
            base_oscillation = base_oscillations[anomaly.get(PARAMETERS.CHANNEL, default_values[ANOMALIES][PARAMETERS.CHANNEL])][PARAMETERS.KIND]
            for anomaly_kind in anomaly.get(PARAMETERS.KINDS, []):
                anomaly_kind = anomaly_kind[PARAMETERS.KIND]
                if not Compatibility.check(anomaly_kind, base_oscillation):
                    if self.skip_errors:
                        logging.warning(f"Skip generation of time series {ts.get('name', '')} due to incompatible types: {anomaly_kind} -> {base_oscillation}.")
                        return False
                    else:
                        raise ValueError(f"Incompatible types: {anomaly_kind} -> {base_oscillation}.")
        return True

    def _skip_name(self, name: str) -> bool:
        return self.only is not None and name != self.only

    def _build_base_oscillations(self, d: Dict) -> List[BaseOscillationInterface]:
        length = d.get(PARAMETERS.LENGTH, default_values[BASE_OSCILLATIONS][PARAMETERS.LENGTH])
        bos = d.get(BASE_OSCILLATIONS, [d.get(BASE_OSCILLATION)] * d.get(PARAMETERS.CHANNELS, 0))
        return [self._build_single_base_oscillation(bo, length) for bo in bos]

    def _build_single_base_oscillation(self, d: Dict, length: int) -> BaseOscillationInterface:
        base_oscillation_config = deepcopy(d)
        base_oscillation_config[PARAMETERS.LENGTH] = length
        trend = base_oscillation_config.get(PARAMETERS.TREND, {})
        base_oscillation_config[PARAMETERS.TREND] = decode_trend_obj(trend, length)
        key = base_oscillation_config[PARAMETERS.KIND]
        return BaseOscillation.from_key(key, **base_oscillation_config)

    def _build_anomalies(self, d: Dict) -> List[Anomaly]:
        return [self._build_single_anomaly(anomaly_config) for anomaly_config in d.get(ANOMALIES, [])]

    def _build_single_anomaly(self, d: Dict) -> Anomaly:
        anomaly = Anomaly(
            Position(d.get(PARAMETERS.POSITION, default_values[ANOMALIES][PARAMETERS.POSITION])),
            d.get(PARAMETERS.EXACT_POSITION, None),
            d[PARAMETERS.LENGTH],
            d.get(PARAMETERS.CHANNEL, default_values[ANOMALIES][PARAMETERS.CHANNEL]),
            d.get(PARAMETERS.CREEP_LENGTH, default_values[ANOMALIES][PARAMETERS.CREEP_LENGTH])
        )

        anomaly_kinds = self._build_anomaly_kinds(d, anomaly.anomaly_length)
        for kind in anomaly_kinds:
            anomaly.set_anomaly(kind)

        return anomaly

    def _build_anomaly_kinds(self, d: Dict, length: int) -> List[BaseAnomaly]:
        return [self._build_single_anomaly_kind(anomaly_kind, length) for anomaly_kind in d.get(PARAMETERS.KINDS, [])]

    def _build_single_anomaly_kind(self, d: Dict, length: int) -> BaseAnomaly:
        kind = d[PARAMETERS.KIND]
        if kind == PARAMETERS.TREND:
            parameters = {
                PARAMETERS.TREND: decode_trend_obj(deepcopy(d[PARAMETERS.OSCILLATION]), length)
            }
        else:
            parameters = deepcopy(d)
            del parameters[PARAMETERS.KIND]
        try:
            return AnomalyKind(kind).create(parameters)
        except TypeError as ex:
            if "unexpected keyword argument" in str(ex):
                parameter = str(ex).split("'")[-2]
                raise ValueError(f"Anomaly kind '{kind}' does not support parameter '{parameter}'") from ex
            else:
                raise ex



# type: ignore  # mypy ends up in recursion

from __future__ import annotations

from enum import Enum
from typing import Optional, List, Dict, Any, NamedTuple, Union

import numpy as np

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext


BASE = "base"
OPERATION = "operation"
OPERAND = "operand"
AGGREGATION = "aggregation"
KIND = "kind"
AXIS = "axis"


class Formula(BaseOscillationInterface):
    def get_timeseries_periods(self) -> Optional[int]:
        return None

    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.Formula

    def generate_only_base(self, ctx: BOGenerationContext, *args, **kwargs) -> np.ndarray:
        formula = self.formula
        c = ctx.previous_channels if ctx.previous_channels else []

        ts = FormulaParser(formula).parse(c).execute()

        return ts


class OperationType(Enum):
    Plus = "+"
    Minus = "-"
    Multiply = "*"
    Divide = "/"
    Power = "**"

    def __call__(self, operand_a: np.ndarray, operand_b: np.ndarray) -> np.ndarray:
        return eval(f"operand_a {self.value} operand_b")


class AggregationType(Enum):
    Sum = "sum"
    Min = "min"
    Max = "max"
    Median = "median"
    Std = "std"
    Var = "var"

    def __call__(self, base: np.ndarray, axis: Optional[int] = None) -> np.ndarray:
        return eval(f"base.{self.value}(axis=axis)")


class Operation(NamedTuple):
    kind: OperationType
    operand: Union[float, FormulaObj]

    def _get_operand(self) -> np.ndarray:
        if type(self.operand) == float:
            return np.array(self.operand)
        return self.operand.execute()

    def execute(self, base: np.ndarray) -> np.ndarray:
        return self.kind(base, self._get_operand())

    @staticmethod
    def from_dict(d: Dict, prev_channels: List[np.ndarray]) -> Operation:
        kind = OperationType(d.get(KIND))
        operand = d.get(OPERAND)

        if type(operand) == float:
            operand = operand
        elif type(operand) == dict:
            operand = FormulaObj.from_dict(operand, prev_channels)
        else:
            raise ValueError("The Operand in Operation has to be either `float` or an `object`")
        return Operation(
            kind=kind,
            operand=operand
        )


class Aggregation(NamedTuple):
    kind: AggregationType
    axis: Optional[int] = None

    def execute(self, base: np.ndarray) -> np.ndarray:
        return self.kind(base, self.axis)

    @staticmethod
    def from_dict(d: Dict) -> Aggregation:
        kind = AggregationType(d.get(KIND))
        axis = d.get(AXIS, None)
        return Aggregation(
            kind=kind,
            axis=axis
        )


class FormulaObj(NamedTuple):
    base: Union[int, FormulaObj]
    prev_channels: List[np.ndarray]
    operation: Optional[Operation] = None
    aggregation: Optional[Aggregation] = None

    def _get_base(self) -> np.ndarray:
        if type(self.base) == int:
            return self.prev_channels[self.base]
        else:
            return self.base.execute()

    def execute(self) -> np.ndarray:
        base = self._get_base()
        if self.operation is not None:
            return self.operation.execute(base)
        elif self.aggregation is not None:
            return self.aggregation.execute(base)
        else:
            return base

    @staticmethod
    def from_dict(d: Dict, prev_channels: List[np.ndarray]) -> FormulaObj:
        base = d.get(BASE)
        operation = d.get(OPERATION, None)
        aggregation = d.get(AGGREGATION, None)
        assert operation is None or aggregation is None, "Only one `operation` or `aggregation` can be set, not both!"

        if type(base) == dict:
            base = FormulaObj.from_dict(base, prev_channels=prev_channels)
        elif type(base) != int:
            raise ValueError("Base must be `float` or `object`.")

        if operation is not None:
            if type(operation) == dict:
                return FormulaObj(
                    base=base,
                    prev_channels=prev_channels,
                    operation=Operation.from_dict(operation, prev_channels)
                )
            else:
                raise ValueError("The Operation has to be an `object`.")
        elif aggregation is not None:
            if type(aggregation) == dict:
                return FormulaObj(
                    base=base,
                    prev_channels=prev_channels,
                    aggregation=Aggregation.from_dict(aggregation)
                )
            else:
                raise ValueError("The Aggregation has to be an `object`.")
        else:
            return FormulaObj(base=base, prev_channels=prev_channels)


class FormulaParser:
    def __init__(self, formula: Dict[str, Any]):
        self.formula = formula

    def parse(self, prev_channels: List[np.ndarray]) -> FormulaObj:
        return FormulaObj.from_dict(self.formula, prev_channels)



from __future__ import annotations

import importlib
import os
from dataclasses import dataclass, field
from typing import List, Type, Optional, Any, Dict

from ..generator import Overview, TimeSeries


@dataclass
class AddOnProcessContext:
    timeseries: TimeSeries
    config: Dict
    plot: bool = False
    output_folder: Optional[os.PathLike] = None

    _data_store: Dict[str, Any] = field(default_factory=lambda: {})

    @property
    def should_save(self) -> bool:
        return self.output_folder is not None

    def copy(self, timeseries: Optional[TimeSeries] = None, config: Optional[Dict] = None) -> AddOnProcessContext:
        return AddOnProcessContext(
            timeseries=timeseries or self.timeseries,
            config=config or self.config,
            plot=self.plot,
            output_folder=self.output_folder,
        )

    def store_data(self, key: str, data: Any) -> AddOnProcessContext:
        ctx = self.copy()
        ctx._data_store[key] = data
        return ctx


@dataclass
class AddOnFinalizeContext:
    overview: Overview
    plot: bool = False
    output_folder: Optional[os.PathLike] = None

    _data_store: Dict[str, List[Any]] = field(default_factory=lambda: {})

    @property
    def should_save(self) -> bool:
        return self.output_folder is not None

    def get_data(self, key: str) -> List[Any]:
        return self._data_store[key]

    def fill_store(self, stores: List[Dict[str, Any]]) -> None:
        """**Don't use!** Internal API."""
        for dd in stores:
            for key in dd:
                if key not in self._data_store:
                    self._data_store[key] = []
                self._data_store[key].append(dd[key])


class BaseAddOn:
    def process(self, ctx: AddOnProcessContext) -> AddOnProcessContext:
        """Gets called after a time series is generated but before they are plotted or written to disk."""
        return ctx

    def finalize(self, ctx: AddOnFinalizeContext) -> None:
        pass


def import_addons(addons: List[str]) -> List[Type[BaseAddOn]]:
    builtin_module = "gutenTAG.addons.builtin"
    module_classes = [(addon.rsplit(".", 1) if "." in addon else (builtin_module, addon)) for addon in addons]

    def load_addon(package, cls):
        try:
            module = importlib.import_module(package)
        except ImportError as ex:
            raise ValueError(f"Package '{package}' for AddOn {cls} could not be loaded!") from ex

        try:
            addon_cls = module.__dict__[cls]
        except KeyError:
            raise ValueError(f"AddOn {cls} not found in package '{package}'!")

        if not issubclass(addon_cls, BaseAddOn):
            raise ValueError(f"Trying to load addon {package}.{cls}, but it is not a compatible AddOn! GutenTAG "
                             "AddOns must inherit from gutenTAG.addons.BaseAddOn!")

        return addon_cls

    return [load_addon(package, cls) for package, cls in module_classes]



import abc
import os
from pathlib import Path
from typing import Dict

import yaml

from ..utils.global_variables import CONFIG_SCHEMA


class ConfigSchemaLoader(abc.ABC):
    @abc.abstractmethod
    def load_schema_file(self, schema_id: str) -> Dict:
        raise NotImplementedError()


class FileSystemConfigSchemaLoader(ConfigSchemaLoader):

    def __init__(self, base_path: Path = CONFIG_SCHEMA.SCHEMA_FOLDER_PATH):
        super().__init__()
        self.base_path = base_path

    def load_schema_file(self, schema_id: str) -> Dict:
        schema_path = CONFIG_SCHEMA.schema_path(schema_id, path=self.base_path)
        with schema_path.open("r") as fh:
            schema = yaml.safe_load(fh)
        return schema

    @staticmethod
    def from_packaged_schema() -> "ConfigSchemaLoader":
        current_path = Path(os.path.dirname(__file__)).absolute()
        return FileSystemConfigSchemaLoader(current_path / "schema")

# We could add a second method of loading the schema using the python package structure similar to
# https://github.com/pallets/jinja/blob/c3a61d6ef654f389ea2bdeddce0ffc74d656be8b/src/jinja2/loaders.py#L238

# class PackageConfigSchemaLoader(ConfigSchemaLoader):
#   pass



from dataclasses import dataclass
from typing import Type

import numpy as np
from scipy.stats import norm
from sklearn.preprocessing import MinMaxScaler

from . import BaseAnomaly, AnomalyProtocol


@dataclass
class AnomalyTrendParameters:
    trend: 'BaseOscillationInterface'  # type: ignore # otherwise we have a circular import


class AnomalyTrend(BaseAnomaly):
    def __init__(self, parameters: AnomalyTrendParameters):
        super().__init__()
        self.trend = parameters.trend

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        length = anomaly_protocol.end - anomaly_protocol.start
        transition_length = int(length * 0.2)
        plateau_length = int(length * 0.8)
        start_transition = norm.pdf(np.linspace(-3, 0, transition_length), scale=1.05)
        amplitude_bell = np.concatenate([start_transition / start_transition.max(), np.ones(plateau_length)])
        amplitude_bell = MinMaxScaler(feature_range=(0, 1)).fit_transform(amplitude_bell.reshape(-1, 1)).reshape(-1)

        self.trend.length = length
        self.trend.generate_timeseries_and_variations(anomaly_protocol.ctx.to_bo())
        timeseries = self.trend.timeseries

        timeseries *= amplitude_bell
        end_point = timeseries[-1]

        anomaly_protocol.base_oscillation.trend_series[anomaly_protocol.start:anomaly_protocol.end] += timeseries
        anomaly_protocol.base_oscillation.trend_series[anomaly_protocol.end:] += end_point

        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return False

    @staticmethod
    def get_parameter_class() -> Type[AnomalyTrendParameters]:
        return AnomalyTrendParameters



from typing import Dict, Optional

import jsonschema
from jsonschema import RefResolver

from ..anomalies import AnomalyKind
from ..base_oscillations import BaseOscillation
from ..config.schema_loader import ConfigSchemaLoader, FileSystemConfigSchemaLoader
from ..utils.global_variables import CONFIG_SCHEMA, TIMESERIES, PARAMETERS, ANOMALIES, BASE_OSCILLATION, \
    BASE_OSCILLATIONS


class GutenTAGParseError(BaseException):
    def __init__(self, prefix: str = "", msg: Optional[str] = None):
        if msg is None:
            msg = prefix
            prefix = ""
        if prefix:
            prefix = f" {prefix}"
        super(GutenTAGParseError, self).__init__(f"Error in generation configuration{prefix}: {msg}")


class ConfigValidator:
    def __init__(self):
        loader: ConfigSchemaLoader = FileSystemConfigSchemaLoader.from_packaged_schema()
        # load base schema
        base_schema_name = CONFIG_SCHEMA.schema_name(CONFIG_SCHEMA.BASE_ID)
        base_schema = loader.load_schema_file(CONFIG_SCHEMA.BASE_ID)

        # load schema parts
        schema_parts = {}
        for schema_part_id in CONFIG_SCHEMA.SCHEMA_PART_IDS:
            name = CONFIG_SCHEMA.schema_name(schema_part_id)
            schema_parts[name] = loader.load_schema_file(schema_part_id)

        # create resolver containing all schema parts
        self.base_schema = base_schema
        self.resolver = RefResolver(
            base_uri=base_schema_name,
            referrer=base_schema,
            store=schema_parts
        )

    def validate(self, config: Dict) -> None:
        self.gutentag_validate(config)
        jsonschema.validate(config, self.base_schema, resolver=self.resolver)

    @staticmethod
    def gutentag_validate(config: Dict) -> None:
        if TIMESERIES not in config:
            raise GutenTAGParseError(f"Key '{TIMESERIES}' not found in root object.")

        for t, ts in enumerate(config.get(TIMESERIES, [])):
            name = ts.get(PARAMETERS.NAME, f"ts_{t}")
            log_prefix = f"TS {name}"

            if ANOMALIES not in ts:
                raise GutenTAGParseError(log_prefix, f"Missing '{ANOMALIES}' property.")

            if BASE_OSCILLATION not in ts and BASE_OSCILLATIONS not in ts:
                raise GutenTAGParseError(log_prefix, f"Missing '{BASE_OSCILLATIONS}' property.")

            if BASE_OSCILLATION in ts and PARAMETERS.CHANNELS not in ts:
                raise GutenTAGParseError(
                    log_prefix,
                    f"If a single '{BASE_OSCILLATION}' is defined, the property '{PARAMETERS.CHANNELS}' is required."
                )

            # check base oscillations
            bos = ts.get(BASE_OSCILLATIONS, [ts.get(BASE_OSCILLATION)] * ts.get(PARAMETERS.CHANNELS, 0))
            for i, bo in enumerate(bos):
                log_prefix_bo = f"{log_prefix} BO {i}"
                if PARAMETERS.KIND not in bo:
                    raise GutenTAGParseError(log_prefix_bo, f"Missing required property '{PARAMETERS.KIND}'.")
                bo_kind = bo[PARAMETERS.KIND]
                if bo_kind not in BaseOscillation.key_mapping:
                    raise GutenTAGParseError(log_prefix_bo, f"Base oscillation kind '{bo_kind}' is not supported!")

            # check anomaly definitions
            anoms = ts.get(ANOMALIES, [])
            for i, anom in enumerate(anoms):
                log_prefix_anom = f"{log_prefix} Anom {i}"
                if PARAMETERS.KINDS not in anom:
                    raise GutenTAGParseError(log_prefix_anom, f"Missing required property '{PARAMETERS.KINDS}'.")
                if PARAMETERS.LENGTH not in anom:
                    raise GutenTAGParseError(log_prefix_anom, f"Missing required property '{PARAMETERS.LENGTH}'.")

                kinds = anom.get(PARAMETERS.KINDS, [])
                for j, anom_kind in enumerate(kinds):
                    log_prefix_kind = f"{log_prefix_anom} Kind {j}"
                    if PARAMETERS.KIND not in anom_kind:
                        raise GutenTAGParseError(log_prefix_kind, f"Missing required property '{PARAMETERS.KIND}'.")
                    if not AnomalyKind.has_value(anom_kind[PARAMETERS.KIND]):
                        raise GutenTAGParseError(
                            log_prefix_kind,
                            f"Anomaly kind '{anom_kind[PARAMETERS.KIND]}' is not supported!"
                        )



from dataclasses import dataclass
from typing import Type

import numpy as np

from . import BaseAnomaly
from .. import AnomalyProtocol
from ...utils.base_oscillation_kind import BaseOscillationKind


@dataclass
class AnomalyPatternShiftParameters:
    shift_by: int = 5
    transition_window: int = 10


class AnomalyPatternShift(BaseAnomaly):
    def __init__(self, parameters: AnomalyPatternShiftParameters):
        super().__init__()
        self.shift_by = parameters.shift_by
        self.transition_window = parameters.transition_window

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        if anomaly_protocol.base_oscillation_kind in [BaseOscillationKind.Sine, BaseOscillationKind.ECG]:
            assert abs(self.shift_by) <= self.transition_window, \
                "The parameter 'shift_by' must not be larger than 'transition_window' in absolute terms! Guten Tag!"

            base = anomaly_protocol.base_oscillation

            subsequence = base.timeseries[anomaly_protocol.start:anomaly_protocol.end]
            transition_start = np.interp(np.linspace(0, self.transition_window, self.transition_window + self.shift_by),
                                         np.arange(self.transition_window), subsequence[:self.transition_window])
            shifted = subsequence[self.transition_window:-self.transition_window]
            transition_end = np.interp(np.linspace(0, self.transition_window, self.transition_window - self.shift_by),
                                       np.arange(self.transition_window), subsequence[-self.transition_window:])

            subsequence = np.concatenate([
                transition_start,
                shifted,
                transition_end
            ])

            anomaly_protocol.subsequences.append(subsequence)
        else:
            self.logger.warn_false_combination(self.__class__.__name__, anomaly_protocol.base_oscillation_kind.name)
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return True

    @staticmethod
    def get_parameter_class() -> Type[AnomalyPatternShiftParameters]:
        return AnomalyPatternShiftParameters



from dataclasses import dataclass
from typing import Type

import numpy as np

from . import BaseAnomaly, AnomalyProtocol


@dataclass
class AnomalyPlatformParameters:
    value: float = 0.0


class AnomalyPlatform(BaseAnomaly):
    def __init__(self, parameters: AnomalyPlatformParameters):
        super().__init__()
        self.value = parameters.value

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        length = anomaly_protocol.end - anomaly_protocol.start
        values = np.zeros(length) + self.value
        anomaly_protocol.subsequences.append(values)
        self.turn_off_trend(anomaly_protocol)
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return False

    @staticmethod
    def get_parameter_class() -> Type[AnomalyPlatformParameters]:
        return AnomalyPlatformParameters



from typing import Optional, Sequence, Callable

import numpy as np

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext


class CylinderBellFunnel(BaseOscillationInterface):
    def get_timeseries_periods(self) -> Optional[int]:
        if self.avg_pattern_length > 0:
            return self.length // self.avg_pattern_length
        return None

    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.CylinderBellFunnel

    def generate_only_base(self,
                           ctx: BOGenerationContext,
                           length: Optional[int] = None,
                           variance: Optional[float] = None,
                           amplitude: Optional[float] = None,
                           variance_pattern_length: Optional[int] = None,
                           *args, **kwargs) -> np.ndarray:
        length = length or self.length
        variance = variance or self.variance
        amplitude = amplitude or self.amplitude
        variance_pattern_length = variance_pattern_length or self.variance_pattern_length

        cbf = generate_pattern_data(
            ctx,
            length=length,
            avg_pattern_length=self.avg_pattern_length,
            avg_amplitude=amplitude,
            default_variance=variance,
            variance_pattern_length=variance_pattern_length,
            variance_amplitude=self.variance_amplitude
        )
        return cbf

    def generate_timeseries_and_variations(self, ctx: BOGenerationContext, **kwargs) -> BaseOscillationInterface:
        super().generate_timeseries_and_variations(ctx)
        if self.timeseries is not None and self.noise is not None:
            self.timeseries -= self.noise
        else:
            raise AssertionError("`timeseries` and `noise` are None. Please, generate `timeseries` and `noise` before calling this method!")
        return self


# Taken from https://github.com/KDD-OpenSource/data-generation/blob/master/generation/cbf.py
# cylinder bell funnel based on "Learning comprehensible descriptions of multivariate time series"
def generate_pattern_data(ctx: BOGenerationContext, length, avg_pattern_length, avg_amplitude, default_variance=1,
                          variance_pattern_length=10, variance_amplitude=2, include_negatives=True):
    def generate_bell(n: int, a: float, v: float) -> np.ndarray:
        bell = ctx.rng.normal(0, v, n) + a * np.arange(n) / n
        return bell

    def generate_funnel(n: int, a: float, v: float) -> np.ndarray:
        funnel = ctx.rng.normal(0, v, n) + a * np.arange(n)[::-1] / n
        return funnel

    def generate_cylinder(n: int, a: float, v: float) -> np.ndarray:
        cylinder = ctx.rng.normal(0, v, n) + a
        return cylinder

    generators: Sequence[Callable[[int, float, float], np.ndarray]] = (generate_bell, generate_funnel, generate_cylinder)
    data = ctx.rng.normal(0, default_variance, length)
    current_start = ctx.rng.integers(0, avg_pattern_length)
    current_length = max(1, int(np.ceil(ctx.rng.normal(avg_pattern_length, variance_pattern_length))))

    while current_start + current_length < length:
        generator: Callable[[int, float, float], np.ndarray] = ctx.rng.choice(generators)  # type: ignore  # strange numpy type prevents chosing a callable
        current_amplitude = ctx.rng.normal(avg_amplitude, variance_amplitude)

        while current_length <= 0:
            current_length = -(current_length - 1)
        pattern = generator(current_length, current_amplitude, default_variance)

        if include_negatives and ctx.rng.random() > 0.5:
            pattern = -1 * pattern

        data[current_start: current_start + current_length] = pattern

        current_start = current_start + current_length + ctx.rng.integers(0, avg_pattern_length)
        current_length = max(1, int(np.ceil(ctx.rng.normal(avg_pattern_length, variance_pattern_length))))

    return data



from abc import ABC, abstractmethod
from typing import Optional

import numpy as np

from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.default_values import default_values
from ..utils.global_variables import PARAMETERS, BASE_OSCILLATIONS
from ..utils.types import BOGenerationContext


class BaseOscillationInterface(ABC):
    def __init__(self, *args, **kwargs):
        self.length = kwargs.get(PARAMETERS.LENGTH, default_values[BASE_OSCILLATIONS][PARAMETERS.LENGTH])
        self.frequency = kwargs.get(PARAMETERS.FREQUENCY, default_values[BASE_OSCILLATIONS][PARAMETERS.FREQUENCY])
        self.amplitude = kwargs.get(PARAMETERS.AMPLITUDE, default_values[BASE_OSCILLATIONS][PARAMETERS.AMPLITUDE])
        self.variance = kwargs.get(PARAMETERS.VARIANCE, default_values[BASE_OSCILLATIONS][PARAMETERS.VARIANCE])
        self.avg_pattern_length = kwargs.get(PARAMETERS.AVG_PATTERN_LENGTH, default_values[BASE_OSCILLATIONS][PARAMETERS.AVG_PATTERN_LENGTH])
        self.variance_pattern_length = kwargs.get(PARAMETERS.VARIANCE_PATTERN_LENGTH, default_values[BASE_OSCILLATIONS][PARAMETERS.VARIANCE_PATTERN_LENGTH])
        self.variance_amplitude = kwargs.get(PARAMETERS.VARIANCE_AMPLITUDE, default_values[BASE_OSCILLATIONS][PARAMETERS.VARIANCE_AMPLITUDE])
        self.freq_mod = kwargs.get(PARAMETERS.FREQ_MOD, default_values[BASE_OSCILLATIONS][PARAMETERS.FREQ_MOD])
        self.polynomial = kwargs.get(PARAMETERS.POLYNOMIAL, default_values[BASE_OSCILLATIONS][PARAMETERS.POLYNOMIAL])
        self.trend: Optional[BaseOscillationInterface] = kwargs.get(PARAMETERS.TREND, default_values[BASE_OSCILLATIONS][PARAMETERS.TREND])
        self.offset = kwargs.get(PARAMETERS.OFFSET, default_values[BASE_OSCILLATIONS][PARAMETERS.OFFSET])
        self.smoothing = kwargs.get(PARAMETERS.SMOOTHING, default_values[BASE_OSCILLATIONS][PARAMETERS.SMOOTHING])
        self.channel_diff = kwargs.get(PARAMETERS.CHANNEL_DIFF, default_values[BASE_OSCILLATIONS][PARAMETERS.CHANNEL_DIFF])
        self.channel_offset = kwargs.get(PARAMETERS.CHANNEL_OFFSET, self.amplitude)
        self.random_seed = kwargs.get(PARAMETERS.RANDOM_SEED, default_values[BASE_OSCILLATIONS][PARAMETERS.RANDOM_SEED])
        self.formula = kwargs.get(PARAMETERS.FORMULA, default_values[BASE_OSCILLATIONS][PARAMETERS.FORMULA])
        self.ecg_sim_method = kwargs.get(PARAMETERS.ECG_SIM_METHOD, default_values[BASE_OSCILLATIONS][PARAMETERS.ECG_SIM_METHOD])

        self.timeseries: Optional[np.ndarray] = None
        self.noise: Optional[np.ndarray] = None
        self.trend_series: Optional[np.ndarray] = None

    def generate_noise(self, ctx: BOGenerationContext, variance: float, length: int) -> np.ndarray:
        return ctx.rng.normal(0, variance, length)

    def _generate_trend(self, ctx: BOGenerationContext) -> np.ndarray:
        trend_series = np.zeros(self.length)
        if self.trend:
            self.trend.length = self.length
            self.trend.generate_timeseries_and_variations(ctx)
            if self.trend.timeseries is not None:
                trend_series = self.trend.timeseries
        return trend_series

    def generate_timeseries_and_variations(self, ctx: BOGenerationContext, **kwargs):
        self.timeseries = self.generate_only_base(ctx, **kwargs)
        self.trend_series = self._generate_trend(ctx.to_trend())
        self.noise = self.generate_noise(ctx, self.variance * self.amplitude, self.length)

    @abstractmethod
    def get_timeseries_periods(self) -> Optional[int]:
        """
        How many same-sized periods do occur in the time series? If no periodicity is given, return None!
        :return: Optional[int]
        """
        raise NotImplementedError()

    @abstractmethod
    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        raise NotImplementedError()

    @abstractmethod
    def generate_only_base(self, ctx: BOGenerationContext, *args, **kwargs) -> np.ndarray:
        raise NotImplementedError()

    @classmethod
    def __subclasshook__(cls, C):
        if cls is BaseOscillationInterface:
            if any("generate" in B.__dict__ for B in C.__mro__):
                return True
        return NotImplemented



from dataclasses import dataclass
from typing import Type

import numpy as np
from sklearn.preprocessing import MinMaxScaler

from . import BaseAnomaly
from .. import AnomalyProtocol
from ...utils.base_oscillation_kind import BaseOscillationKind


@dataclass
class AnomalyPatternParameters:
    sinusoid_k: float = 10.0
    cbf_pattern_factor: float = 2.0


class AnomalyPattern(BaseAnomaly):
    def __init__(self, parameters: AnomalyPatternParameters):
        super().__init__()
        self.sinusoid_k = parameters.sinusoid_k
        self.cbf_pattern_factor = parameters.cbf_pattern_factor

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        if anomaly_protocol.base_oscillation_kind == BaseOscillationKind.Sine:
            def sinusoid(t: np.ndarray, k: float, a_min: float, a_max: float) -> np.ndarray:
                pattern = (np.arctan(k * t) / np.arctan(k))
                scaled = MinMaxScaler(feature_range=(a_min, a_max)).fit_transform(pattern.reshape(-1, 1)).reshape(-1)
                return scaled

            sine = anomaly_protocol.base_oscillation
            snippet = sine.timeseries[anomaly_protocol.start:anomaly_protocol.end]
            subsequence = sinusoid(snippet, self.sinusoid_k, snippet.min(), snippet.max())
            anomaly_protocol.subsequences.append(subsequence)

        elif anomaly_protocol.base_oscillation_kind == BaseOscillationKind.CylinderBellFunnel:
            cbf = anomaly_protocol.base_oscillation
            subsequence = cbf.generate_only_base(
                anomaly_protocol.ctx.to_bo(),
                variance_pattern_length=cbf.variance_pattern_length * self.cbf_pattern_factor
            )[anomaly_protocol.start:anomaly_protocol.end]
            anomaly_protocol.subsequences.append(subsequence)

        elif anomaly_protocol.base_oscillation_kind == BaseOscillationKind.ECG:
            ecg = anomaly_protocol.base_oscillation
            length = anomaly_protocol.end - anomaly_protocol.start
            window = int(length * 0.05)

            for slide in range(-3, 3):
                start = ecg.timeseries[anomaly_protocol.start+slide:anomaly_protocol.start+window]
                if np.argmax(start) == 0:
                    break
            else:
                slide = 0

            subsequence = ecg.timeseries[anomaly_protocol.start + slide:anomaly_protocol.end + slide][::-1]
            anomaly_protocol.subsequences.append(subsequence)

        else:
            self.logger.warn_false_combination(self.__class__.__name__, anomaly_protocol.base_oscillation_kind.name)
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return True

    @staticmethod
    def get_parameter_class() -> Type[AnomalyPatternParameters]:
        return AnomalyPatternParameters



from __future__ import annotations

import json
import os
import warnings
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Optional, Union, Callable, Sequence, Tuple, Any

import yaml
from joblib import Parallel, delayed
from tqdm import tqdm

from .addons import import_addons, AddOnProcessContext, AddOnFinalizeContext, BaseAddOn
from .config.parser import ConfigParser
from .config.validator import ConfigValidator
from .generator.overview import Overview
from .generator.timeseries import TimeSeries
from .timeseries import TrainingType, TimeSeries as ExtTimeSeries
from .utils.global_variables import UNSUPERVISED_FILENAME, SUPERVISED_FILENAME, SEMI_SUPERVISED_FILENAME
from .utils.tqdm_joblib import tqdm_joblib


@dataclass
class _GenerationContext:
    plot: bool = False
    return_timeseries: bool = False
    output_folder: Optional[os.PathLike] = None
    seed: Optional[int] = None
    addons: Sequence[BaseAddOn] = ()

    def to_addon_process_ctx(self, timeseries: TimeSeries, config: Dict) -> AddOnProcessContext:
        return AddOnProcessContext(
            timeseries=timeseries,
            config=config,
            plot=self.plot,
            output_folder=self.output_folder,
        )


class GutenTAG:
    def __init__(self,
                 n_jobs: int = 1,
                 seed: Optional[int] = None,
                 addons: Sequence[str] = ()):
        self._overview = Overview()
        self._timeseries: List[TimeSeries] = []
        self._n_jobs = n_jobs
        # remove duplicate addons
        self._registered_addons: List[str] = []
        for addon in addons:
            self.use_addon(addon)
        self._overview.add_seed(seed)
        self.seed = seed
        self.addons: Dict[str, BaseAddOn] = {}

    def load_config_json(self, json_config_path: os.PathLike, only: Optional[str] = None) -> GutenTAG:
        with open(json_config_path, "r") as f:
            config = json.load(f)
        return self.load_config_dict(config, only)

    def load_config_yaml(self, yaml_config_path: os.PathLike, only: Optional[str] = None) -> GutenTAG:
        with open(yaml_config_path, "r") as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
        return self.load_config_dict(config, only)

    def load_config_dict(self, config: Dict, only: Optional[str] = None) -> GutenTAG:
        # First parse, then validate config, because our own error messages are more precise than
        # the validator's ones.
        config_parser = ConfigParser(only=only)
        timeseries = []
        for base_oscillations, anomalies, options, ts_config in config_parser.parse(config):
            ts = TimeSeries(base_oscillations, anomalies, **options.to_dict())
            timeseries.append(ts)
        ConfigValidator().validate(config)

        self._timeseries.extend(timeseries)
        self._overview.add_datasets(config_parser.raw_ts_configs)
        return self

    def remove_by_name(self, name: Union[str, Callable[[str], bool]]) -> GutenTAG:
        if isinstance(name, str):
            self._timeseries = [ts for ts in self._timeseries if ts.dataset_name != name]
        else:
            self._timeseries = [ts for ts in self._timeseries if not name(ts.dataset_name)]
        self._overview.remove_dataset_by_name(name)
        return self

    def use_addon(self, addon: str, insert_location: Union[str, int] = "last") -> GutenTAG:
        if addon not in self._registered_addons:
            if insert_location == "last":
                self._registered_addons.append(addon)
            elif insert_location == "first":
                self._registered_addons = [addon] + self._registered_addons
            elif isinstance(insert_location, int):
                self._registered_addons.insert(insert_location, addon)
            else:
                ValueError(f"insert_position={insert_location} unknown!")
        else:
            ValueError(f"'{addon}' already loaded at position {self._registered_addons.index(addon)}")
        return self

    def generate(self,
                 return_timeseries: bool = False,
                 output_folder: Optional[os.PathLike] = None,
                 plot: bool = False) -> Optional[List[ExtTimeSeries]]:
        n_jobs = self._n_jobs
        if n_jobs != 1 and plot:
            warnings.warn(
                f"Cannot generate time series in parallel while plotting ('n_jobs' was set to {n_jobs})! Falling "
                f"back to serial generation.")
            n_jobs = 1

        # prepare
        folder: Optional[Path] = None
        if output_folder is not None:
            folder = Path(output_folder)
            folder.mkdir(exist_ok=True)

        addon_types = import_addons(list(self._registered_addons))
        addons = [addon() for addon in tqdm(addon_types, desc="Initializing addons", total=len(addon_types))]
        for name, addon in zip(self._registered_addons, addons):
            self.addons[name] = addon

        # process time series
        ctx = _GenerationContext(
            plot=plot,
            output_folder=output_folder,
            seed=self.seed,
            addons=addons,
            return_timeseries=return_timeseries,
        )
        with tqdm_joblib(tqdm(desc="Generating datasets", total=len(self._timeseries))):
            results: List[Tuple[Dict, Dict[str, Any], Optional[List[ExtTimeSeries]]]] = Parallel(n_jobs=n_jobs)(
                delayed(self.internal_generate)(ctx, ts, config)
                for ts, config in zip(self._timeseries, self._overview.datasets)
            )
        configs, data_dicts, timeseries_datasets = tuple(zip(*results))
        self._overview.datasets = configs

        # finalize
        if folder is not None:
            self._overview.save_to_output_dir(folder)
        finalize_ctx = AddOnFinalizeContext(
            overview=self._overview,
            plot=plot,
            output_folder=output_folder
        )
        finalize_ctx.fill_store(data_dicts)
        for addon in tqdm(addons, desc="Finalizing addons", total=len(addons)):
            addon.finalize(finalize_ctx)

        if return_timeseries and timeseries_datasets is not None:
            return [d for dataset in timeseries_datasets for d in dataset]
        return None

    @staticmethod
    def internal_generate(ctx: _GenerationContext, ts: TimeSeries, config: Dict) -> Tuple[Dict, Dict[str, Any], Optional[List[ExtTimeSeries]]]:
        ts.generate(ctx.seed)
        addon_ctx = ctx.to_addon_process_ctx(ts, config)
        for addon in ctx.addons:
            addon_ctx = addon.process(addon_ctx)
        ts = addon_ctx.timeseries
        config = addon_ctx.config
        data = addon_ctx._data_store

        if ctx.plot:
            ts.plot()

        if ctx.output_folder is not None:
            GutenTAG.save_timeseries(ts, ctx.output_folder)

        if ctx.return_timeseries:
            return config, data, ts.to_datasets()
        return config, data, None

    @staticmethod
    def save_timeseries(ts: TimeSeries, output_dir: os.PathLike) -> None:
        name = ts.dataset_name
        dataset_folder = Path(output_dir) / name
        dataset_folder.mkdir(exist_ok=True)

        ts.to_csv(dataset_folder / UNSUPERVISED_FILENAME, TrainingType.TEST)

        if ts.supervised:
            ts.to_csv(dataset_folder / SUPERVISED_FILENAME, TrainingType.TRAIN_ANOMALIES)

        if ts.semi_supervised:
            ts.to_csv(dataset_folder / SEMI_SUPERVISED_FILENAME, TrainingType.TRAIN_NO_ANOMALIES)

    @staticmethod
    def from_json(path: os.PathLike,
                  n_jobs: int = 1,
                  seed: Optional[int] = None,
                  addons: Sequence[str] = (),
                  only: Optional[str] = None) -> GutenTAG:
        gt = GutenTAG(n_jobs=n_jobs, seed=seed, addons=addons)
        return gt.load_config_json(path, only=only)

    @staticmethod
    def from_yaml(path: os.PathLike,
                  n_jobs: int = 1,
                  seed: Optional[int] = None,
                  addons: Sequence[str] = (),
                  only: Optional[str] = None) -> GutenTAG:
        gt = GutenTAG(n_jobs=n_jobs, seed=seed, addons=addons)
        return gt.load_config_yaml(path, only=only)

    @staticmethod
    def from_dict(config: Dict,
                  n_jobs: int = 1,
                  seed: Optional[int] = None,
                  addons: Sequence[str] = (),
                  only: Optional[str] = None) -> GutenTAG:
        gt = GutenTAG(n_jobs=n_jobs, seed=seed, addons=addons)
        return gt.load_config_dict(config, only=only)



import argparse
import sys
import warnings
from pathlib import Path
from typing import List

from ._version import __version__
from .gutenTAG import GutenTAG


def parse_args(args: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="GutenTAG! A good Time series Anomaly Generator.")

    parser.add_argument("--version", action="store_true", help="Display GutenTAG version and exit.")
    parser.add_argument("--config-yaml", type=Path, required=True, help="Path to config YAML")
    parser.add_argument("--output-dir", type=Path, default=Path("./generated-timeseries"),
                        help="Path to output directory")
    parser.add_argument("--plot", action="store_true", help="Plot every generated time series.")
    parser.add_argument("--no-save", action="store_true",
                        help="Prevent GutenTAG from saving the generated time series.")
    parser.add_argument("--seed", type=int, default=None, help="Random seed")
    parser.add_argument("--addons", nargs="*", default=[], help="Use Add-Ons for generating time series.")
    parser.add_argument("--n_jobs", "--n-jobs", type=int, default=1,
                        help="Number of time series to generate in parallel.")
    parser.add_argument("--only", type=str, help="Process only timeseries with the defined name.")

    return parser.parse_args(args)


def main(sys_args: List[str]) -> None:
    print(f"""

                      Welcome to

       _____       _          _______       _____ _
      / ____|     | |        |__   __|/\\   / ____| |
     | |  __ _   _| |_ ___ _ __ | |  /  \\ | |  __| |
     | | |_ | | | | __/ _ \\ '_ \\| | / /\\ \\| | |_ | |
     | |__| | |_| | ||  __/ | | | |/ ____ \\ |__| |_|
      \\_____|\\__,_|\\__\\___|_| |_|_/_/    \\_\\_____(_)

                     Version {__version__}

"Good day!" wishes your friendly Timeseries Anomaly Generator.


""")

    if "--version" in sys_args:
        return

    args = parse_args(sys_args)
    if args.no_save:
        output = None
    else:
        output = Path(args.output_dir)

    gutentag = GutenTAG.from_yaml(
        args.config_yaml,
        n_jobs=args.n_jobs,
        seed=args.seed,
        addons=args.addons,
        only=args.only
    )

    gutentag.generate(
        return_timeseries=False,
        output_folder=output,
        plot=args.plot
    )


def cli() -> None:
    main(sys.argv[1:])


if __name__ == "__main__":
    cli()



from typing import Optional

import numpy as np

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext


class Sine(BaseOscillationInterface):
    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.Sine

    def get_timeseries_periods(self) -> Optional[int]:
        return int((self.length / 100) * self.frequency)

    def generate_only_base(self,
                           ctx: BOGenerationContext,
                           length: Optional[int] = None,
                           frequency: Optional[float] = None,
                           amplitude: Optional[float] = None,
                           freq_mod: Optional[float] = None,
                           *args, **kwargs) -> np.ndarray:
        v_length: int = length or self.length
        v_frequency: float = frequency or self.frequency
        v_amplitude: float = amplitude or self.amplitude
        v_freq_mod: float = freq_mod or self.freq_mod

        periods = (v_length / 100) * v_frequency

        end = 2 * np.pi * periods
        base_ts = np.linspace(0, end, v_length)

        arr_amplitude: np.ndarray = np.array(v_amplitude)
        if v_freq_mod:
            arr_amplitude = np.sin(np.linspace(0, end * v_freq_mod, v_length)) * v_amplitude

        return np.sin(base_ts) * arr_amplitude



from __future__ import annotations

from hashlib import md5
from pathlib import Path
from typing import Optional, List, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from numpy.random import SeedSequence

from ..anomalies import Anomaly
from ..base_oscillations import BaseOscillationInterface
from ..base_oscillations.utils.consolidator import Consolidator
from ..timeseries import TrainingType, INDEX_COLUMN_NAME, LABEL_COLUMN_NAME, TimeSeries as ExtTimeSeries
from ..utils.types import GenerationContext


class TimeSeries:
    def __init__(self, base_oscillations: List[BaseOscillationInterface], anomalies: List[Anomaly],
                 dataset_name: str, semi_supervised: bool = False, supervised: bool = False):
        self.dataset_name = dataset_name
        self.base_oscillations = base_oscillations
        self.anomalies = anomalies
        self.semi_supervised_timeseries: Optional[np.ndarray] = None
        self.supervised_timeseries: Optional[np.ndarray] = None
        self.timeseries: Optional[np.ndarray] = None
        self.labels: Optional[np.ndarray] = None
        self.train_labels: Optional[np.ndarray] = None
        self.semi_train_labels: Optional[np.ndarray] = None
        self.semi_supervised = semi_supervised
        self.supervised = supervised
        self._rng_counter = 0

    def generate(self, random_seed: Optional[int] = None) -> TimeSeries:
        consolidator = Consolidator(self.base_oscillations, self.anomalies)
        self.timeseries, self.labels = consolidator.generate(GenerationContext(seed=self._create_new_seed(random_seed)))

        if self.semi_supervised:
            semi_supervised_consolidator = Consolidator(self.base_oscillations, [])
            self.semi_supervised_timeseries, self.semi_train_labels = semi_supervised_consolidator.generate(
                GenerationContext(seed=self._create_new_seed(random_seed))
            )

        if self.supervised:
            supervised_consolidator = Consolidator(self.base_oscillations, self.anomalies)
            self.supervised_timeseries, self.train_labels = supervised_consolidator.generate(
                GenerationContext(seed=self._create_new_seed(random_seed))
            )

        return self

    def generate_with_dataframe(self, random_seed: Optional[int] = None) -> pd.DataFrame:
        self.generate(random_seed)
        return self.to_dataframe()

    def plot(self) -> None:
        n_series = 1 + np.sum([self.semi_supervised, self.supervised])
        fig, axs = plt.subplots(2, n_series, sharex="col", sharey="row", figsize=(6*n_series, 5))
        # fix indexing, because subplots only returns a 1-dim array in this case:
        if n_series == 1:
            axs = np.array([axs]).T

        fig.suptitle(self.dataset_name)

        names: List[str] = ["test"]
        assert self.timeseries is not None, "TimeSeries is not generated. Please, do so before plotting!"
        assert self.labels is not None, "TimeSeries is not generated. Please, do so before plotting!"
        series: List[np.ndarray] = [self.timeseries]
        labels: List[np.ndarray] = [self.labels]
        if self.supervised and self.supervised_timeseries is not None and self.train_labels is not None:
            names.append("train_supervised")
            series.append(self.supervised_timeseries)
            labels.append(self.train_labels)
        if self.semi_supervised and self.semi_supervised_timeseries is not None and self.semi_train_labels is not None:
            names.append("train_semi-supervised")
            series.append(self.semi_supervised_timeseries)
            labels.append(self.semi_train_labels)
        for i, (name, ts, label) in enumerate(zip(names, series, labels)):
            axs[0, i].set_title(name)
            name_list = list(map(lambda j: f"channel-{j}", range(ts.shape[1]))) if ts.shape[1] > 1 else "time series"
            axs[0, i].plot(ts, label=name_list)
            axs[1, i].plot(label, color="orange", label="ground truth")
        axs[0, 0].legend()
        axs[1, 0].legend()
        plt.show()

    def build_figure_base_oscillation(self) -> plt.Figure:
        assert self.timeseries is not None, "TimeSeries is not generated. Please, do so before building a figure!"

        channels = self.timeseries.shape[1]
        name = list(map(lambda j: f"channel-{j}", range(channels))) if channels > 1 else "time series"

        fig, ax = plt.subplots()
        ax.plot(self.timeseries, label=name)
        return fig

    def to_datasets(self) -> List[ExtTimeSeries]:
        results: List[ExtTimeSeries] = []
        training_types = [TrainingType.TEST]
        if self.semi_supervised:
            training_types.append(TrainingType.TRAIN_NO_ANOMALIES)
        if self.supervised:
            training_types.append(TrainingType.TRAIN_ANOMALIES)

        for training_type in training_types:
            results.append(ExtTimeSeries(
                name=self.dataset_name,
                training_type=training_type,
                timeseries=self.to_dataframe(training_type)
            ))
        return results

    def to_dataframe(self, training_type: TrainingType = TrainingType.TEST) -> pd.DataFrame:
        if training_type == TrainingType.TEST:
            ts, labels = self.timeseries, self.labels
        elif training_type == TrainingType.TRAIN_NO_ANOMALIES:
            ts, labels = self.semi_supervised_timeseries, self.semi_train_labels
        else:  # if training_type == TrainingType.TRAIN_ANOMALIES:
            ts, labels = self.supervised_timeseries, self.train_labels

        assert ts is not None, f"The timeseries for {training_type.value} must be generated before creating a DataFrame"
        if labels is None:
            labels = np.zeros(ts.shape[0])
        channel_names = list(map(lambda i: f"value-{i}", range(ts.shape[1])))
        df = pd.DataFrame(ts, columns=channel_names)
        df.index.name = INDEX_COLUMN_NAME
        df[LABEL_COLUMN_NAME] = labels
        return df

    def to_csv(self, output_dir: Path, training_type: TrainingType = TrainingType.TEST) -> None:
        df = self.to_dataframe(training_type)
        df.to_csv(output_dir, sep=",", index=True)

    def _create_new_seed(self, base_seed: Optional[int]) -> SeedSequence:
        if base_seed is None:
            base_seed1: Union[int, SeedSequence] = SeedSequence()
        else:
            base_seed1 = base_seed
        seeds = [int.from_bytes(md5(self.dataset_name.encode("utf-8")).digest(), byteorder="big")]
        if self._rng_counter > 0:
            seeds.append(self._rng_counter)
        self._rng_counter += 1
        return GenerationContext.re_seed(seeds, base_seed1)



import os
from typing import List, Dict, Optional, Any, Callable, Union

import git
import numpy as np
import yaml


class DictSanitizer:
    NUMPY_TYPES = tuple(list(np.core._type_aliases.allTypes.values()) + [np.ndarray])  # type: ignore # mypy does not find allTypes

    def sanitize(self, obj: Dict) -> Dict:
        for key, value in obj.items():
            obj[key] = self._sanitize_value(value)
        return obj

    def _decode_numpy_types(self, obj: Any) -> Any:
        """
        Taken from [numpyencoder](https://github.com/hmallen/numpyencoder/blob/f8199a61ccde25f829444a9df4b21bcb2d1de8f2/numpyencoder/numpyencoder.py)
        """

        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)

        elif isinstance(obj, (np.float_, np.float16, np.float32, np.float64)):
            return float(obj)

        elif isinstance(obj, (np.complex_, np.complex64, np.complex128)):
            return {'real': obj.real, 'imag': obj.imag}

        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()

        elif isinstance(obj, (np.bool_)):
            return bool(obj)

        elif isinstance(obj, (np.void)):
            return None

    def _sanitize_value(self, obj: Any) -> Any:
        if isinstance(obj, dict):
            obj = self.sanitize(obj)  # type: ignore # recursive typecheck
        elif isinstance(obj, list):
            obj = list(map(self._sanitize_value, obj))  # type: ignore # recursive typecheck
        elif isinstance(obj, self.NUMPY_TYPES):
            obj = self._decode_numpy_types(obj)
        return obj


class Overview:
    FILENAME = "overview.yaml"
    GUTENTAG_LINK = "https://gitlab.hpi.de/akita/guten-tag"

    def __init__(self):
        self.datasets: List[Dict] = []
        self.seed: Optional[int] = None

        try:
            self.git_commit_sha = git.Repo(search_parent_directories=True).head.object.hexsha
        except git.InvalidGitRepositoryError:
            self.git_commit_sha = None

    def add_seed(self, seed: Optional[int]) -> None:
        self.seed = seed

    def add_dataset(self, config: Dict) -> None:
        self.datasets.append(config)

    def add_datasets(self, configs: List[Dict]) -> None:
        self.datasets.extend(configs)

    def remove_dataset_by_name(self, name: Union[str, Callable[[str], bool]]) -> None:
        if isinstance(name, str):
            self.datasets = [d for d in self.datasets if d["name"] != name]
        else:
            self.datasets = [d for d in self.datasets if not name(d["name"])]

    def save_to_output_dir(self, path: os.PathLike) -> None:
        overview: Dict[str, Any] = dict()
        overview["generated-timeseries"] = []
        for i, dataset in enumerate(self.datasets):
            dataset["generation-id"] = dataset.get("base_oscillation", {}).get("title", i)
            overview["generated-timeseries"].append(dataset)

        overview["meta"] = {}
        overview["meta"]["seed"] = self.seed
        overview["meta"]["git_commit_sha"] = self.git_commit_sha
        overview["meta"]["download_link"] = self.GUTENTAG_LINK

        overview = DictSanitizer().sanitize(overview)

        with open(os.path.join(path, self.FILENAME), "w") as f:
            yaml.dump(overview, f)



import glob
import os
import shutil
import sys
from pathlib import Path

from setuptools import setup, find_packages, Command


HERE = Path(os.path.dirname(__file__)).absolute()
# get __version__ from gutenTAG/_version.py
with open(HERE / "gutenTAG" / "_version.py") as f:
    exec(f.read())
VERSION: str = __version__  # noqa
README = (HERE / "README.md").read_text(encoding="UTF-8")
DOC_NAME = "GutenTAG"
PYTHON_NAME = "gutenTAG"
with open(HERE / "requirements.txt") as fh:
    REQUIRED = fh.read().splitlines()


class PyTestCommand(Command):
    description = f"run PyTest for {DOC_NAME}"
    user_options = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        import pytest
        from pytest import ExitCode

        exit_code = pytest.main(["--cov-report=term", "--cov-report=xml:coverage.xml",
                                 f"--cov={PYTHON_NAME}", "tests"])
        if exit_code == ExitCode.TESTS_FAILED:
            raise ValueError("Tests failed!")
        elif exit_code == ExitCode.INTERRUPTED:
            raise ValueError("pytest was interrupted!")
        elif exit_code == ExitCode.INTERNAL_ERROR:
            raise ValueError("pytest internal error!")
        elif exit_code == ExitCode.USAGE_ERROR:
            raise ValueError("Pytest was not correctly used!")
        elif exit_code == ExitCode.NO_TESTS_COLLECTED:
            raise ValueError("No tests found!")
        # else: everything is fine


class MyPyCheckCommand(Command):
    description = f"run MyPy for {DOC_NAME}; performs static type checking"
    user_options = []

    def initialize_options(self) -> None:
        pass

    def finalize_options(self) -> None:
        pass

    def run(self) -> None:
        from mypy.main import main as mypy

        args = ["--pretty", PYTHON_NAME, "tests"]
        mypy(None, stdout=sys.stdout, stderr=sys.stderr, args=args)


class CleanCommand(Command):
    description = "Remove build artifacts from the source tree"
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        files = [
            ".coverage*",
            "coverage.xml"
        ]
        dirs = ["build", "dist", "*.egg-info", "**/__pycache__", ".mypy_cache",
                ".pytest_cache", "**/.ipynb_checkpoints"]
        for d in dirs:
            for filename in glob.glob(d):
                shutil.rmtree(filename, ignore_errors=True)

        for f in files:
            for filename in glob.glob(f):
                try:
                    os.remove(filename)
                except OSError:
                    pass


if __name__ == "__main__":
    setup(
        name=f"timeeval-{PYTHON_NAME}",
        version=VERSION,
        description="A good Timeseries Anomaly Generator.",
        long_description=README,
        long_description_content_type="text/markdown",
        author="Phillip Wenig and Sebastian Schmidl",
        author_email="phillip.wenig@hpi.de",
        url="https://github.com/HPI-Information-Systems/gutentag",
        license="MIT",
        classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10"
        ],
        packages=find_packages(exclude=("tests", "tests.*")),
        package_data={"gutenTAG": ["py.typed", "config/schema/*"]},
        install_requires=REQUIRED,
        python_requires=">=3.7",
        test_suite="tests",
        cmdclass={
            "test": PyTestCommand,
            "typecheck": MyPyCheckCommand,
            "clean": CleanCommand
        },
        zip_safe=False,
        # provides="gutenTAG",
        entry_points={
            "console_scripts": [
                "gutenTAG=gutenTAG.__main__:cli"
            ]
        }
    )



import pandas as pd
from ..utils.global_variables import BASE_OSCILLATION_NAMES, ANOMALY_TYPE_NAMES


class Compatibility:
    combinations = pd.DataFrame(
        [
            [1, 1, 1, 1, 0, 0, 0],  # amplitude
            [1, 1, 1, 1, 1, 0, 1],  # extremum
            [1, 0, 0, 1, 0, 0, 0],  # frequency
            [1, 1, 1, 1, 1, 0, 1],  # mean
            [1, 0, 1, 1, 0, 0, 0],  # pattern
            [1, 0, 0, 1, 0, 0, 0],  # pattern_shift
            [1, 1, 1, 1, 1, 0, 1],  # platform
            [1, 1, 1, 1, 1, 0, 1],  # trend
            [1, 1, 1, 1, 1, 0, 1],  # variance
            [0, 0, 0, 0, 0, 1, 0],  # mode_correlation
        ],
        columns=[BASE_OSCILLATION_NAMES.SINE, BASE_OSCILLATION_NAMES.RANDOM_WALK, BASE_OSCILLATION_NAMES.CYLINDER_BELL_FUNNEL,
                 BASE_OSCILLATION_NAMES.ECG, BASE_OSCILLATION_NAMES.POLYNOMIAL, BASE_OSCILLATION_NAMES.RANDOM_MODE_JUMP,
                 BASE_OSCILLATION_NAMES.FORMULA],
        index=[ANOMALY_TYPE_NAMES.AMPLITUDE, ANOMALY_TYPE_NAMES.EXTREMUM, ANOMALY_TYPE_NAMES.FREQUENCY, ANOMALY_TYPE_NAMES.MEAN,
               ANOMALY_TYPE_NAMES.PATTERN, ANOMALY_TYPE_NAMES.PATTERN_SHIFT, ANOMALY_TYPE_NAMES.PLATFORM, ANOMALY_TYPE_NAMES.TREND,
               ANOMALY_TYPE_NAMES.VARIANCE, ANOMALY_TYPE_NAMES.MODE_CORRELATION],
        dtype=bool
    )

    @staticmethod
    def check(anomaly: str, base_oscillation: str) -> bool:
        return Compatibility.combinations.loc[anomaly, base_oscillation]



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import List, Optional

import numpy as np

from ...utils.base_oscillation_kind import BaseOscillationKind
from ...utils.logger import GutenTagLogger
from ...utils.types import AnomalyGenerationContext


@dataclass
class LabelRange:
    start: int
    length: int


@dataclass
class AnomalyProtocol:
    start: int
    end: int
    channel: int
    ctx: AnomalyGenerationContext
    labels: LabelRange
    creep_length: int = 0
    subsequences: List[np.ndarray] = field(default_factory=lambda: [])

    @property
    def rng(self) -> np.random.Generator:
        return self.ctx.rng

    @property
    def base_oscillation(self) -> 'BaseOscillationInterface':  # type: ignore # otherwise we have a circular import
        return self.ctx.base_oscillation

    @property
    def base_oscillation_kind(self) -> BaseOscillationKind:
        return self.ctx.base_oscillation_kind

    @property
    def length(self) -> int:
        return self.end - self.start

    @property
    def length_without_creep(self) -> int:
        return self.length - self.creep_length


class BaseAnomaly(ABC):
    def __init__(self):
        self.logger = GutenTagLogger()

    @abstractmethod
    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        return anomaly_protocol

    @property
    @abstractmethod
    def requires_period_start_position(self) -> bool:
        return False

    def generate_creep(self, anomaly_protocol: AnomalyProtocol, custom_anomaly_length: Optional[int] = None) -> np.ndarray:
        creep_length = anomaly_protocol.creep_length
        anomaly_length = anomaly_protocol.length_without_creep if custom_anomaly_length is None else custom_anomaly_length
        return np.concatenate([
            np.linspace(0, 1, creep_length),  # creep
            np.ones(anomaly_length)           # anomaly
        ])

    def turn_off_trend(self, anomaly_protocol):
        anomaly_protocol.base_oscillation.trend_series[anomaly_protocol.start:anomaly_protocol.end] = 0

    @staticmethod
    @abstractmethod
    def get_parameter_class():
        raise NotImplementedError()



from dataclasses import dataclass
from enum import Enum

import pandas as pd


class TrainingType(Enum):
    TEST = "test"
    TRAIN_NO_ANOMALIES = "train-no-anomaly"
    TRAIN_ANOMALIES = "train-anomaly"


@dataclass
class TimeSeries:
    name: str
    training_type: TrainingType
    timeseries: pd.DataFrame


INDEX_COLUMN_NAME = "timestamp"
LABEL_COLUMN_NAME = "is_anomaly"



from typing import Optional, Any, Dict, Type

from .cylinder_bell_funnel import CylinderBellFunnel
from .ecg import ECG
from .formula import Formula  # type: ignore  # mypy ends up in recursion
from .interface import BaseOscillationInterface
from .polynomial import Polynomial
from .random_mode_jump import RandomModeJump
from .random_walk import RandomWalk
from .sine import Sine
from ..utils.global_variables import BASE_OSCILLATION_NAMES


def get_or_error(name: str, value: Optional[Any]) -> Any:
    if value is None:
        raise ValueError(f"Parameter {name} for the base_oscillation must be set!")
    return value


class BaseOscillation:
    key_mapping: Dict[str, Type[BaseOscillationInterface]] = {
        BASE_OSCILLATION_NAMES.SINE: Sine,
        BASE_OSCILLATION_NAMES.RANDOM_WALK: RandomWalk,
        BASE_OSCILLATION_NAMES.CYLINDER_BELL_FUNNEL: CylinderBellFunnel,
        BASE_OSCILLATION_NAMES.ECG: ECG,
        BASE_OSCILLATION_NAMES.POLYNOMIAL: Polynomial,
        BASE_OSCILLATION_NAMES.RANDOM_MODE_JUMP: RandomModeJump,
        BASE_OSCILLATION_NAMES.FORMULA: Formula
    }

    @staticmethod
    def from_key(key: str, *args, **kwargs) -> BaseOscillationInterface:
        return BaseOscillation.key_mapping[key](*args, **kwargs)



from dataclasses import dataclass
from typing import Type

from . import BaseAnomaly, AnomalyProtocol
from ...utils.base_oscillation_kind import BaseOscillationKind


@dataclass
class AnomalyFrequencyParameters:
    frequency_factor: float = 1.0


class AnomalyFrequency(BaseAnomaly):
    def __init__(self, parameters: AnomalyFrequencyParameters):
        super().__init__()
        self.frequency_factor = parameters.frequency_factor

    def generate(self, anomaly_protocol: AnomalyProtocol) -> AnomalyProtocol:
        if anomaly_protocol.base_oscillation_kind == BaseOscillationKind.Sine:
            sine = anomaly_protocol.base_oscillation
            length = anomaly_protocol.end - anomaly_protocol.start
            subsequence = sine.generate_only_base(anomaly_protocol.ctx.to_bo(), length, sine.frequency * self.frequency_factor, freq_mod=sine.freq_mod)
            anomaly_protocol.subsequences.append(subsequence)
        elif anomaly_protocol.base_oscillation_kind == BaseOscillationKind.ECG:
            ecg = anomaly_protocol.base_oscillation
            subsequence = ecg.generate_only_base(anomaly_protocol.ctx.to_bo(), frequency=ecg.frequency * self.frequency_factor)[anomaly_protocol.start:anomaly_protocol.end]
            anomaly_protocol.subsequences.append(subsequence)
        else:
            self.logger.warn_false_combination(self.__class__.__name__, anomaly_protocol.base_oscillation_kind.name)
        return anomaly_protocol

    @property
    def requires_period_start_position(self) -> bool:
        return True

    @staticmethod
    def get_parameter_class() -> Type[AnomalyFrequencyParameters]:
        return AnomalyFrequencyParameters



from typing import Optional

import numpy as np

from .interface import BaseOscillationInterface
from ..utils.base_oscillation_kind import BaseOscillationKind
from ..utils.types import BOGenerationContext


class RandomModeJump(BaseOscillationInterface):
    def get_base_oscillation_kind(self) -> BaseOscillationKind:
        return BaseOscillationKind.RandomModeJump

    def get_timeseries_periods(self) -> Optional[int]:
        return self.frequency

    def generate_only_base(self,
                           ctx: BOGenerationContext,
                           length: Optional[int] = None,
                           frequency: Optional[float] = None,
                           channel_diff: Optional[float] = None,
                           channel_offset: Optional[float] = None,
                           random_seed: Optional[int] = None,
                           *args, **kwargs) -> np.ndarray:
        length = length or self.length
        frequency = frequency or self.frequency
        channel_diff = channel_diff or self.channel_diff
        channel_offset = channel_offset or self.channel_offset
        random_seed = random_seed or self.random_seed
        step_length = int(length // frequency)

        base_random_steps = self._generate_random_steps(ctx, length, step_length, random_seed)
        channel_amplitude = self._generate_channel_amplitude(ctx.channel, channel_diff, channel_offset)
        ts = base_random_steps * channel_amplitude

        return ts

    def _generate_random_steps(self, ctx: BOGenerationContext, length: int, step_length: int, random_seed: Optional[int] = None) -> np.ndarray:
        rng = np.random.default_rng(ctx.re_seed(random_seed, base_seed=ctx.seed))
        n_steps = int(np.ceil(length / step_length).item())
        energy = rng.choice([-1, 1], size=n_steps)
        base = np.repeat(energy, step_length)[:length].astype(np.float64)
        return base

    def _generate_channel_amplitude(self, channel: int, channel_diff: float, channel_offset: float) -> float:
        high_val = (channel_diff * channel) + channel_offset
        return high_val
